import asyncio
import base64
import hashlib
import io
import time
from typing import TYPE_CHECKING, Any, Set, cast

from lonelypsp.stateful.constants import BroadcasterToSubscriberStatefulMessageType
from lonelypsp.stateful.messages.confirm_configure import B2S_ConfirmConfigureParser
from lonelypsp.stateful.parser_helpers import parse_b2s_message_prefix
from lonelypsp.util.bounded_deque import BoundedDeque
from lonelypsp.util.drainable_asyncio_queue import DrainableAsyncioQueue

from lonelypsc.client import (
    PubSubCancelRequested,
    PubSubError,
    PubSubIrrecoverableError,
)
from lonelypsc.types.websocket_message import WSMessageBytes
from lonelypsc.ws.check_result import (
    CheckResult,
    CheckStateChangerResult,
    CheckStateChangerResultContinue,
    CheckStateChangerResultDone,
)
from lonelypsc.ws.compressor import CompressorStoreImpl
from lonelypsc.ws.handlers.protocol import StateHandler
from lonelypsc.ws.handlers.util.read_from_websocket import make_websocket_read_task
from lonelypsc.ws.handlers.util.state_specific_cleanup import handle_via_composition
from lonelypsc.ws.state import (
    ClosingRetryInformationCannotRetry,
    ClosingRetryInformationType,
    ClosingRetryInformationWantRetry,
    ManagementTask,
    ManagementTaskSubscribeExact,
    ManagementTaskSubscribeGlob,
    ManagementTaskType,
    OpenRetryInformationTentative,
    OpenRetryInformationType,
    SendingSimple,
    SendingState,
    State,
    StateClosing,
    StateConfiguring,
    StateOpen,
    StateType,
)


async def handle_configuring(state: State) -> State:
    """Waits for the broadcaster to respond with the confirm configure message,
    then moves to the OPEN state

    If there are errors, handles them in the same way as in CONNECTING
    """
    assert state.type == StateType.CONFIGURING
    return await handle_via_composition(
        state, core=_core, recover=_recover, shutdown=_shutdown
    )


async def _core(state: StateConfiguring) -> State:
    """
    Happy path to progress configuring state, raises an error if cleanup code
    needs to be called, typically PubSubError for recoverable errors and
    PubSubIrrecoverableError for unrecoverable errors
    """
    if await _check_send_task(state) == CheckResult.RESTART:
        return state

    if (result := await _check_read_task(state)).type == CheckResult.RESTART:
        return result.state

    if (result := await _check_cancel_requested(state)).type == CheckResult.RESTART:
        return result.state

    if (result := _check_backgrounded(state)).type == CheckResult.RESTART:
        return result.state

    wait_cancel_requested = asyncio.create_task(state.cancel_requested.wait())
    await asyncio.wait(
        [
            state.read_task,
            *([state.send_task] if state.send_task is not None else []),
            wait_cancel_requested,
            *state.backgrounded,
        ],
        return_when=asyncio.FIRST_COMPLETED,
    )
    wait_cancel_requested.cancel()
    return state


async def _check_send_task(state: StateConfiguring) -> CheckResult:
    """Checks if the subscriber is done with the application part of sending the
    CONFIGURE message
    """
    if state.send_task is None or not state.send_task.done():
        return CheckResult.CONTINUE

    task = state.send_task
    state.send_task = None
    task.result()
    return CheckResult.RESTART


async def _check_read_task(state: StateConfiguring) -> CheckStateChangerResult:
    """Checks if the broadcaster has confirmed the configure message yet"""
    if not state.read_task.done():
        return CheckStateChangerResultContinue(type=CheckResult.CONTINUE)

    raw_message = state.read_task.result()
    if raw_message["type"] == "websocket.disconnect":
        raise Exception("disconnected before confirming configure")

    if "bytes" not in raw_message:
        raise Exception("received non-bytes non-disconnect message")

    message = cast(WSMessageBytes, raw_message)
    payload = message["bytes"]

    stream = io.BytesIO(payload)
    prefix = parse_b2s_message_prefix(stream)

    if prefix.type != BroadcasterToSubscriberStatefulMessageType.CONFIRM_CONFIGURE:
        raise PubSubError(
            f"received unexpected message before confirm configure: {prefix}"
        )

    parsed_message = B2S_ConfirmConfigureParser.parse(prefix.flags, prefix.type, stream)

    auth_result = await state.config.is_stateful_confirm_configure_allowed(
        message=parsed_message, now=time.time()
    )
    if auth_result != "ok":
        raise PubSubError(f"confirm configure authorization failed: {auth_result}")

    connection_nonce = hashlib.sha256(
        state.subscriber_nonce + parsed_message.broadcaster_nonce
    ).digest()

    management_tasks: DrainableAsyncioQueue[ManagementTask] = DrainableAsyncioQueue(
        max_size=state.config.max_expected_acks
    )
    for topic in state.tasks.exact_subscriptions:
        management_tasks.put_nowait(
            ManagementTaskSubscribeExact(
                type=ManagementTaskType.SUBSCRIBE_EXACT, topic=topic
            )
        )
    for glob in state.tasks.glob_subscriptions:
        management_tasks.put_nowait(
            ManagementTaskSubscribeGlob(
                type=ManagementTaskType.SUBSCRIBE_GLOB, glob=glob
            )
        )

    for task in state.tasks.unsorted.drain():
        management_tasks.put_nowait(task)

    return CheckStateChangerResultDone(
        type=CheckResult.RESTART,
        state=StateOpen(
            type=StateType.OPEN,
            client_session=state.client_session,
            config=state.config,
            cancel_requested=state.cancel_requested,
            broadcaster=state.broadcaster,
            broadcaster_counter=1,
            subscriber_counter=-1,
            nonce_b64=base64.b64encode(connection_nonce).decode("ascii"),
            websocket=state.websocket,
            retry=OpenRetryInformationTentative(
                type=OpenRetryInformationType.TENTATIVE,
                stable_at=time.time() + state.config.outgoing_min_reconnect_interval,
                continuation=state.retry,
            ),
            compressors=CompressorStoreImpl(),
            unsent_notifications=state.tasks.unsent_notifications,
            resending_notifications=state.tasks.resending_notifications,
            sent_notifications=BoundedDeque(maxlen=state.config.max_sent_notifications),
            exact_subscriptions=set(),
            glob_subscriptions=set(),
            management_tasks=management_tasks,
            expected_acks=BoundedDeque(maxlen=state.config.max_expected_acks),
            receiving=None,
            received=DrainableAsyncioQueue(max_size=state.config.max_received),
            unsent_acks=BoundedDeque(maxlen=state.config.max_unsent_acks),
            sending=(
                None
                if state.send_task is None
                else SendingSimple(type=SendingState.SIMPLE, task=state.send_task)
            ),
            read_task=make_websocket_read_task(state.websocket),
            backgrounded=state.backgrounded,
        ),
    )


async def _check_cancel_requested(state: StateConfiguring) -> CheckStateChangerResult:
    """Checks if it has been requested that the state move towards CLOSED"""
    if not state.cancel_requested.is_set():
        return CheckStateChangerResultContinue(type=CheckResult.CONTINUE)
    raise PubSubCancelRequested()


def _check_backgrounded(state: StateConfiguring) -> CheckStateChangerResult:
    """Cleans out done background tasks, raising an irrecoverable error if any
    failed
    """
    if not any(task.done() for task in state.backgrounded):
        return CheckStateChangerResultContinue(type=CheckResult.CONTINUE)

    new_backgrounded: Set[asyncio.Task[Any]] = set()
    for task in state.backgrounded:
        if not task.done():
            new_backgrounded.add(task)
            continue

        if task.exception() is not None:
            raise PubSubIrrecoverableError("background task failed")

    state.backgrounded = new_backgrounded
    return CheckStateChangerResultDone(type=CheckResult.RESTART, state=state)


async def _cleanup(state: StateConfiguring) -> None:
    """Cancels any pending tasks; can be called multiple times"""
    if state.send_task is not None:
        state.send_task.cancel()
    state.read_task.cancel()


async def _recover(state: StateConfiguring, /, *, cause: BaseException) -> State:
    """Attempts to move to the next broadcaster"""
    await _cleanup(state)
    return StateClosing(
        type=StateType.CLOSING,
        config=state.config,
        cancel_requested=state.cancel_requested,
        broadcaster=state.broadcaster,
        client_session=state.client_session,
        websocket=state.websocket,
        retry=(
            ClosingRetryInformationWantRetry(
                type=ClosingRetryInformationType.WANT_RETRY,
                retry=state.retry,
                tasks=state.tasks,
                exception=cause,
            )
        ),
        backgrounded=state.backgrounded,
    )


async def _shutdown(state: StateConfiguring, /, *, cause: BaseException) -> State:
    """Moves towards the CLOSED state"""
    await _cleanup(state)
    return StateClosing(
        type=StateType.CLOSING,
        config=state.config,
        cancel_requested=state.cancel_requested,
        broadcaster=state.broadcaster,
        client_session=state.client_session,
        websocket=state.websocket,
        retry=ClosingRetryInformationCannotRetry(
            type=ClosingRetryInformationType.CANNOT_RETRY,
            tasks=state.tasks,
            exception=cause,
        ),
        backgrounded=state.backgrounded,
    )


if TYPE_CHECKING:
    _: StateHandler = handle_configuring



from typing import TYPE_CHECKING

from lonelypsp.stateful.messages.continue_notify import B2S_ContinueNotify

from lonelypsc.client import PubSubError, PubSubIrrecoverableError
from lonelypsc.ws.handlers.open.messages.protocol import MessageChecker
from lonelypsc.ws.state import StateOpen


def check_continue_notify(state: StateOpen, message: B2S_ContinueNotify) -> None:
    """Handles the subscriber receiving the a message from the broadcaster that it
    is ready for more data on a related NOTIFY_STREAM the subscriber previously
    sent, raising an error if this isn't the acknowledgement the subscriber was
    expecting.
    """
    try:
        expected = state.expected_acks.popleft()
    except IndexError:
        raise PubSubError(f"{message.type} received when not expecting any acks")

    if expected.type != message.type:
        raise PubSubError(f"expected {expected.type}, got {message.type}")
    if expected.identifier != message.identifier:
        raise PubSubError(
            f"expected {message.type} {expected.identifier!r}, got {message.identifier!r}"
        )

    try:
        internal_message = state.sent_notifications[0]
    except IndexError:
        raise PubSubIrrecoverableError(
            f"invariant violated: expected {expected.type}, but no corresponding internal message"
        )

    if internal_message.identifier != message.identifier:
        raise PubSubIrrecoverableError(
            f"invariant violated: expected {expected.type} {message.identifier!r}, got {internal_message.identifier!r}"
        )


if TYPE_CHECKING:
    _: MessageChecker[B2S_ContinueNotify] = check_continue_notify



import asyncio
import secrets
from typing import TYPE_CHECKING, List

from lonelypsp.stateful.constants import SubscriberToBroadcasterStatefulMessageType
from lonelypsp.stateful.messages.configure import S2B_Configure, serialize_s2b_configure

from lonelypsc.client import PubSubCancelRequested, PubSubIrrecoverableError
from lonelypsc.util.errors import combine_multiple_exceptions
from lonelypsc.ws.check_result import (
    CheckResult,
    CheckStateChangerResult,
    CheckStateChangerResultContinue,
    CheckStateChangerResultDone,
)
from lonelypsc.ws.handle_connection_failure import (
    cleanup_tasks_and_raise,
    handle_connection_failure,
)
from lonelypsc.ws.handlers.protocol import StateHandler
from lonelypsc.ws.handlers.util.read_from_websocket import make_websocket_read_task
from lonelypsc.ws.handlers.util.state_specific_cleanup import handle_via_composition
from lonelypsc.ws.internal_callbacks import sweep_internal_message
from lonelypsc.ws.state import (
    State,
    StateConfiguring,
    StateConnecting,
    StateType,
)


async def handle_connecting(state: State) -> State:
    """Tries to connect to the given broadcaster; if unsuccessful,
    moves to either CONNECTING (with the next broadcaster),
    WAITING_RETRY, or CLOSED. If successful, moves to CONFIGURING.
    May need multiple calls to progress
    """
    assert state.type == StateType.CONNECTING
    return await handle_via_composition(
        state,
        core=_core,
        recover=_recover,
        shutdown=_shutdown,
    )


async def _core(state: StateConnecting) -> State:
    if (result := await _check_canceled(state)).type == CheckResult.RESTART:
        return result.state

    if (result := await _check_websocket(state)).type == CheckResult.RESTART:
        return result.state

    await _sweep_resending_notifications(state)
    await _sweep_backgrounded(state)
    await _wait_something_changed(state)
    return state


async def _recover(state: StateConnecting, /, *, cause: BaseException) -> State:
    try:
        await state.client_session.close()
    except Exception as e:
        cause = combine_multiple_exceptions(
            "failed to close session", [e], context=cause
        )

    return await handle_connection_failure(
        config=state.config,
        cancel_requested=state.cancel_requested,
        retry=state.retry,
        tasks=state.tasks,
        exception=cause,
        backgrounded=state.backgrounded,
    )


async def _shutdown(state: StateConnecting, /, *, cause: BaseException) -> State:
    cleanup_excs: List[BaseException] = []
    if state.websocket_task.done():
        try:
            websocket = state.websocket_task.result()
            await websocket.close()
        except BaseException as e:
            cleanup_excs.append(e)
    try:
        await state.client_session.close()
    except BaseException as e:
        cleanup_excs.append(e)

    if cleanup_excs:
        cause = combine_multiple_exceptions(
            "failed to close websocket", cleanup_excs, context=cause
        )

    await cleanup_tasks_and_raise(
        state.tasks,
        state.backgrounded,
        "shutdown from connecting",
        cause,
    )


async def _wait_something_changed(state: StateConnecting) -> None:
    wait_cancel = asyncio.create_task(state.cancel_requested.wait())
    await asyncio.wait(
        [
            wait_cancel,
            state.websocket_task,
            *[
                msg.callback.task
                for msg in state.tasks.resending_notifications
                if msg.callback.task is not None
            ],
            *state.backgrounded,
        ],
        return_when=asyncio.FIRST_COMPLETED,
    )
    wait_cancel.cancel()


async def _check_canceled(state: StateConnecting) -> CheckStateChangerResult:
    if not state.cancel_requested.is_set():
        return CheckStateChangerResultContinue(type=CheckResult.CONTINUE)

    raise PubSubCancelRequested()


async def _sweep_backgrounded(state: StateConnecting) -> None:
    if not any(j.done() for j in state.backgrounded):
        return

    new_backgrounded = set()

    for bknd in state.backgrounded:
        if not bknd.done():
            new_backgrounded.add(bknd)
            continue

        if bknd.exception() is None:
            continue

        # avoids duplicating the error as it will be found during cleanup
        # again
        raise PubSubIrrecoverableError("saw backgrounded task failed")

    state.backgrounded = new_backgrounded


async def _sweep_resending_notifications(state: StateConnecting) -> None:
    for msg in state.tasks.resending_notifications:
        sweep_internal_message(msg)


async def _check_websocket(state: StateConnecting) -> CheckStateChangerResult:
    if not state.websocket_task.done():
        return CheckStateChangerResultContinue(type=CheckResult.CONTINUE)

    websocket = state.websocket_task.result()
    subscriber_nonce = secrets.token_bytes(32)
    tracing = b""  # TODO: tracing
    enable_zstd = state.config.allow_compression
    enable_training = state.config.allow_training_compression
    initial_dict = state.config.initial_compression_dict_id or 0
    authorization = await state.config.authorize_stateful_configure(
        tracing=tracing,
        subscriber_nonce=subscriber_nonce,
        enable_zstd=enable_zstd,
        enable_training=enable_training,
        initial_dict=initial_dict,
    )
    return CheckStateChangerResultDone(
        type=CheckResult.RESTART,
        state=StateConfiguring(
            type=StateType.CONFIGURING,
            client_session=state.client_session,
            config=state.config,
            cancel_requested=state.cancel_requested,
            broadcaster=state.broadcaster,
            websocket=websocket,
            retry=state.retry,
            tasks=state.tasks,
            subscriber_nonce=subscriber_nonce,
            send_task=asyncio.create_task(
                websocket.send_bytes(
                    serialize_s2b_configure(
                        S2B_Configure(
                            type=SubscriberToBroadcasterStatefulMessageType.CONFIGURE,
                            subscriber_nonce=subscriber_nonce,
                            enable_zstd=enable_zstd,
                            enable_training=enable_training,
                            initial_dict=initial_dict,
                            authorization=authorization,
                            tracing=tracing,
                        ),
                        minimal_headers=state.config.websocket_minimal_headers,
                    )
                )
            ),
            read_task=make_websocket_read_task(websocket),
            backgrounded=state.backgrounded,
        ),
    )


if TYPE_CHECKING:
    _: StateHandler = handle_connecting



from lonelypsc.client import PubSubIrrecoverableError
from lonelypsc.ws.check_result import CheckResult
from lonelypsc.ws.state import StateOpen


def check_backgrounded(state: StateOpen) -> CheckResult:
    if not any(j.done() for j in state.backgrounded):
        return CheckResult.CONTINUE

    new_backgrounded = set()

    for bknd in state.backgrounded:
        if not bknd.done():
            new_backgrounded.add(bknd)
            continue

        if bknd.exception() is None:
            continue

        # avoids duplicating the error as it will be found during cleanup
        # again
        raise PubSubIrrecoverableError("saw backgrounded task failed")

    state.backgrounded = new_backgrounded
    return CheckResult.RESTART



from enum import Enum, auto
from typing import TYPE_CHECKING, Literal, Union

from lonelypsp.compat import fast_dataclass

if TYPE_CHECKING:
    from lonelypsc.ws.state import State


class CheckResult(Enum):
    """The handler function is composed into parts; each part can either request
    we return back to the top of the handler function, continue to the next part
    of the handler function, or raise an exception to start the cleanup and disconnect
    process.

    Although the return type could be annotated as a boolean, for clarity
    we instead use an enum
    """

    RESTART = auto()
    """Return to the top of the handler function because we made progress"""

    CONTINUE = auto()
    """Continue to the next part of the handler function"""


@fast_dataclass
class CheckStateChangerResultContinue:
    """
    Generally only used as a union type when a check function may switch state
    types, which can never be done concurrently with another function that might
    do so
    """

    type: Literal[CheckResult.CONTINUE]
    """Indicates that nothing was changed and the handler should proceed to the next check,
    if any, otherwise wait for something to change and restart
    """


@fast_dataclass
class CheckStateChangerResultDone:
    """
    Generally only used if the check function might switch states, which can
    never be done concurrently with another function that might do so
    """

    type: Literal[CheckResult.RESTART]
    """Indicates that the check function changed something and the handler should return
    the new state.
    """
    state: "State"


CheckStateChangerResult = Union[
    CheckStateChangerResultContinue, CheckStateChangerResultDone
]
"""The result of a check function that might switch state types rather than just
potentially mutating the state without changing the type
"""



import asyncio
from typing import TYPE_CHECKING, Optional

from lonelypsp.stateful.messages.enable_zstd_preset import B2S_EnableZstdPreset

from lonelypsc.client import PubSubError
from lonelypsc.ws.compressor import (
    CompressorPreparing,
    CompressorReady,
    CompressorState,
)
from lonelypsc.ws.handlers.open.messages.protocol import MessageChecker
from lonelypsc.ws.state import (
    StateOpen,
)

try:
    import zstandard
except ImportError:
    ...


async def _make_compressor(
    state: StateOpen, message: B2S_EnableZstdPreset
) -> CompressorReady:
    zdict: Optional[zstandard.ZstdCompressionDict] = None
    if message.identifier != 1:
        zdict = await state.config.get_compression_dictionary_by_id(
            message.identifier, level=message.compression_level
        )
        if zdict is None:
            raise PubSubError(f"unknown preset dictionary {message.identifier}")
    return CompressorReady(
        type=CompressorState.READY,
        identifier=message.identifier,
        level=message.compression_level,
        min_size=message.min_size,
        max_size=message.max_size,
        data=zdict,
        compressors=list(),
        decompressors=list(),
    )


def check_enable_zstd_preset(state: StateOpen, message: B2S_EnableZstdPreset) -> None:
    """Handles the subscriber receiving the a message from the broadcaster that
    the broadcaster intends to use a specific preset dictionary for compressing
    the messages on this websocket
    """
    if message.identifier == 0:
        return

    if not state.config.allow_compression:
        raise PubSubError("compression is disabled but a preset was requested")

    state.compressors.add_compressor(
        CompressorPreparing(
            type=CompressorState.PREPARING,
            identifier=message.identifier,
            task=asyncio.create_task(_make_compressor(state, message)),
        )
    )


if TYPE_CHECKING:
    _: MessageChecker[B2S_EnableZstdPreset] = check_enable_zstd_preset



from typing import TYPE_CHECKING

from lonelypsc.client import PubSubCancelRequested
from lonelypsc.ws.check_result import CheckResult
from lonelypsc.ws.handlers.open.check_backgrounded import check_backgrounded
from lonelypsc.ws.handlers.open.check_management_tasks import check_management_tasks
from lonelypsc.ws.handlers.open.check_read import check_read_task
from lonelypsc.ws.handlers.open.check_receiving_authorizing import (
    check_receiving_authorizing,
)
from lonelypsc.ws.handlers.open.check_receiving_authorizing_missed import (
    check_receiving_authorizing_missed,
)
from lonelypsc.ws.handlers.open.check_receiving_decompressing import (
    check_receiving_decompressing,
)
from lonelypsc.ws.handlers.open.check_receiving_waiting_compressor import (
    check_receiving_waiting_compressor,
)
from lonelypsc.ws.handlers.open.check_resending_notifications import (
    check_resending_notifications,
)
from lonelypsc.ws.handlers.open.check_retry_tentative import check_retry_tentative
from lonelypsc.ws.handlers.open.check_send import check_sending
from lonelypsc.ws.handlers.open.check_sent_notifications import check_sent_notifications
from lonelypsc.ws.handlers.open.check_unsent_acks import check_unsent_acks
from lonelypsc.ws.handlers.open.check_unsent_notifications import (
    check_unsent_notifications,
)
from lonelypsc.ws.handlers.open.cleanup import recover_open, shutdown_open
from lonelypsc.ws.handlers.open.wait_something_changed import wait_something_changed
from lonelypsc.ws.handlers.protocol import StateHandler
from lonelypsc.ws.handlers.util.state_specific_cleanup import handle_via_composition
from lonelypsc.ws.state import State, StateOpen, StateType


async def handle_open(state: State) -> State:
    assert state.type == StateType.OPEN
    return await handle_via_composition(
        state,
        core=_core,
        recover=recover_open,
        shutdown=shutdown_open,
    )


async def _core(state: StateOpen) -> State:
    """The core inner loop for the websocket client; processes incoming
    messages and sends outgoing messages in a deterministic order
    """
    if state.cancel_requested.is_set():
        raise PubSubCancelRequested()

    if check_receiving_authorizing_missed(state) == CheckResult.RESTART:
        return state
    if check_receiving_authorizing(state) == CheckResult.RESTART:
        return state
    if check_receiving_waiting_compressor(state) == CheckResult.RESTART:
        return state
    if check_receiving_decompressing(state) == CheckResult.RESTART:
        return state
    if check_read_task(state) == CheckResult.RESTART:
        return state
    if check_sending(state) == CheckResult.RESTART:
        return state
    if check_unsent_acks(state) == CheckResult.RESTART:
        return state
    if check_management_tasks(state) == CheckResult.RESTART:
        return state
    if check_resending_notifications(state) == CheckResult.RESTART:
        return state
    if check_unsent_notifications(state) == CheckResult.RESTART:
        return state
    if check_sent_notifications(state) == CheckResult.RESTART:
        return state
    if state.compressors.check_compressor_tasks() == CheckResult.RESTART:
        return state
    if check_backgrounded(state) == CheckResult.RESTART:
        return state
    if check_retry_tentative(state) == CheckResult.RESTART:
        return state
    await wait_something_changed(state)
    return state


if TYPE_CHECKING:
    _: StateHandler = handle_open



from io import BytesIO, FileIO
from typing import TYPE_CHECKING, Protocol, Type, Union


class SyncReadableBytesIOA(Protocol):
    """A type that represents a stream that can be read synchronously"""

    def read(self, n: int) -> bytes:
        """Reads n bytes from the file-like object"""
        raise NotImplementedError()


class SyncReadableBytesIOB(Protocol):
    """A type that represents a stream that can be read synchronously"""

    def read(self, n: int, /) -> bytes:
        """Reads n bytes from the file-like object"""
        raise NotImplementedError()


SyncReadableBytesIO = Union[SyncReadableBytesIOA, SyncReadableBytesIOB]


class SyncTellableBytesIO(Protocol):
    """A type that represents a stream with a synchronous tell method"""

    def tell(self) -> int:
        """Returns the current position in the file"""
        raise NotImplementedError()


class SyncSeekableBytesIO(Protocol):
    """A type that represents a stream with a synchronous seek method"""

    def seek(self, offset: int, whence: int = 0) -> int:
        """Seeks to a position in the file"""
        raise NotImplementedError()


class SyncStandardIOA(
    SyncReadableBytesIOA, SyncTellableBytesIO, SyncSeekableBytesIO, Protocol
): ...


class SyncStandardIOB(
    SyncReadableBytesIOA, SyncTellableBytesIO, SyncSeekableBytesIO, Protocol
): ...


SyncStandardIO = Union[SyncStandardIOA, SyncStandardIOB]


class SyncLengthIO(Protocol):
    """A type that represents a stream with a synchronous length method"""

    def __len__(self) -> int:
        """Returns the length of the file"""
        raise NotImplementedError()


class SyncStandardWithLengthIOA(
    SyncReadableBytesIOA,
    SyncTellableBytesIO,
    SyncSeekableBytesIO,
    SyncLengthIO,
    Protocol,
): ...


class SyncStandardWithLengthIOB(
    SyncReadableBytesIOA,
    SyncTellableBytesIO,
    SyncSeekableBytesIO,
    SyncLengthIO,
    Protocol,
): ...


SyncStandardWithLengthIO = Union[SyncStandardWithLengthIOA, SyncStandardWithLengthIOB]
"""A type that usually only occurs manually, since it doesn't match BytesIO or FileIO,
but is used to represent a stream where the length is known in advance
"""


class SyncWritableBytesIO(Protocol):
    """A type that represents a stream that can be written synchronously"""

    def write(self, b: Union[bytes, bytearray], /) -> int:
        """Writes the given bytes to the file-like object"""
        raise NotImplementedError()


class Closeable(Protocol):
    """Represents something that can be closed"""

    def close(self) -> None:
        """Closes the file-like object"""


class SyncIOBaseLikeIOA(
    SyncReadableBytesIOA,
    SyncTellableBytesIO,
    SyncSeekableBytesIO,
    SyncWritableBytesIO,
    Closeable,
    Protocol,
): ...


class SyncIOBaseLikeIOB(
    SyncReadableBytesIOB,
    SyncTellableBytesIO,
    SyncSeekableBytesIO,
    SyncWritableBytesIO,
    Closeable,
    Protocol,
): ...


SyncIOBaseLikeIO = Union[SyncIOBaseLikeIOA, SyncIOBaseLikeIOB]
"""For when you want to use RawIOBase but it's not a protocol"""

if TYPE_CHECKING:
    # verifies BytesIO matches
    _a: Type[SyncReadableBytesIO] = BytesIO
    _b: Type[SyncTellableBytesIO] = BytesIO
    _c: Type[SyncSeekableBytesIO] = BytesIO
    _d: Type[SyncStandardIO] = BytesIO

    # verifies the result of open matches
    _e: Type[SyncReadableBytesIO] = FileIO
    _f: Type[SyncTellableBytesIO] = FileIO
    _g: Type[SyncSeekableBytesIO] = FileIO
    _h: Type[SyncStandardIO] = FileIO



import asyncio
import hashlib
import io
import os
import secrets
import tempfile
import time
from typing import Optional, Union

from lonelypsp.stateful.constants import (
    BroadcasterToSubscriberStatefulMessageType,
    SubscriberToBroadcasterStatefulMessageType,
)
from lonelypsp.stateful.messages.confirm_notify import B2S_ConfirmNotify
from lonelypsp.stateful.messages.continue_notify import B2S_ContinueNotify
from lonelypsp.stateful.messages.notify import (
    S2B_NotifyCompressed,
    S2B_NotifyUncompressed,
    serialize_s2b_notify,
)
from lonelypsp.stateful.messages.notify_stream import (
    S2B_NotifyStreamContinuation,
    S2B_NotifyStreamStartCompressed,
    S2B_NotifyStreamStartUncompressed,
    serialize_s2b_notify_stream,
)
from lonelypsp.stateful.parser_helpers import read_exact

from lonelypsc.types.sync_io import SyncReadableBytesIO
from lonelypsc.ws.compressor import CompressorReady, CompressorState
from lonelypsc.ws.handlers.open.compressor_utils import reserve_compressor
from lonelypsc.ws.internal_callbacks import (
    inform_internal_message,
    readable_internal_message,
)
from lonelypsc.ws.state import (
    InternalLargeMessage,
    InternalMessage,
    InternalMessageStateSent,
    InternalMessageStateType,
    InternalMessageType,
    InternalSmallMessage,
    StateOpen,
)


async def send_internal_message(
    state: StateOpen, message: Union[InternalSmallMessage, InternalLargeMessage]
) -> None:
    """Expected to be run in the sending.task; sends the given internal
    message over the websocket, adding expected acks as appropriate

    Informs the callback while holding the read lock for the message, and
    queues it to be set to SENT
    """
    async with readable_internal_message(message):
        # will occur only once we exit the context manager
        inform_internal_message(
            message,
            InternalMessageStateSent(
                type=InternalMessageStateType.SENT,
            ),
        )
        if message.type == InternalMessageType.SMALL:
            return await send_internal_small_message(state, message)
        return await send_internal_large_message(state, message)


async def send_internal_small_message(
    state: StateOpen, message: InternalSmallMessage
) -> None:
    """Expected to be run in the sending.task; sends a message which is
    entirely in memory over the websocket, adding expected acks as
    appropriate.

    WARN: does not inform the message callback
    """

    compressor = state.compressors.get_for_compression(len(message.data))
    if compressor is None:
        return await send_internal_small_message_uncompressed(state, message)

    if compressor.type == CompressorState.PREPARING:
        compressor = await compressor.task

    await send_internal_small_message_compressed(state, message, compressor=compressor)


async def send_internal_large_message(
    state: StateOpen, message: InternalLargeMessage
) -> None:
    """Expected to be run in the sending.task; sends a message which is
    too large to hold in memory over the websocket, adding expected acks as
    appropriate.

    Note that this will detect and optimize (reduce python<->c calls) the case
    where the uncompressed payload is too large to hold in memory but the
    compressed payload is small enough

    WARN: does not inform the message callback
    """
    assert message.stream.tell() == 0, "stream must be at start"
    compressor = state.compressors.get_for_compression(message.length)
    if compressor is None:
        return await send_internal_large_message_uncompressed(state, message)

    if compressor.type == CompressorState.PREPARING:
        compressor = await compressor.task

    await send_internal_large_message_compressed(state, message, compressor=compressor)


async def send_internal_small_message_uncompressed(
    state: StateOpen, message: InternalSmallMessage, /
) -> None:
    """Sends the given internal small message uncompressed; will use NOTIFY
    if possible, otherwise NOTIFY_STREAM

    WARN: does not inform the message callback
    """
    tracing = b""  # TODO: tracing
    identifier = secrets.token_bytes(4)
    authorization = await state.config.authorize_notify(
        tracing=tracing,
        topic=message.topic,
        identifier=identifier,
        message_sha512=message.sha512,
        now=time.time(),
    )

    if (
        state.config.max_websocket_message_size is None
        or len(message.data) + 128 < state.config.max_websocket_message_size
    ):
        as_simple_message = serialize_s2b_notify(
            S2B_NotifyUncompressed(
                type=SubscriberToBroadcasterStatefulMessageType.NOTIFY,
                authorization=authorization,
                tracing=tracing,
                identifier=message.identifier,
                compressor_id=None,
                topic=message.topic,
                verified_uncompressed_sha512=message.sha512,
                uncompressed_message=message.data,
            ),
            minimal_headers=state.config.websocket_minimal_headers,
        )
        if (
            state.config.max_websocket_message_size is None
            or len(as_simple_message) < state.config.max_websocket_message_size
        ):
            state.expected_acks.append(
                B2S_ConfirmNotify(
                    type=BroadcasterToSubscriberStatefulMessageType.CONFIRM_NOTIFY,
                    identifier=message.identifier,
                    subscribers=-1,
                    tracing=b"",
                    authorization=None,
                )
            )
            state.sent_notifications.append(message)
            await state.websocket.send_bytes(as_simple_message)
            return

    await send_notify_stream_given_first_headers(
        state=state,
        stream=io.BytesIO(message.data),
        length=len(message.data),
        identifier=message.identifier,
        topic=message.topic,
        sha512=message.sha512,
        first_headers=serialize_s2b_notify_stream(
            S2B_NotifyStreamStartUncompressed(
                type=SubscriberToBroadcasterStatefulMessageType.NOTIFY_STREAM,
                authorization=authorization,
                tracing=tracing,
                identifier=message.identifier,
                part_id=None,
                topic=message.topic,
                compressor_id=None,
                uncompressed_length=len(message.data),
                unverified_uncompressed_sha512=message.sha512,
                payload=b"",
            ),
            minimal_headers=state.config.websocket_minimal_headers,
        ),
        msg=message,
    )


async def send_internal_large_message_uncompressed(
    state: StateOpen,
    message: InternalLargeMessage,
    /,
    *,
    authorization: Optional[str] = None,
) -> None:
    """Sends the given large message uncompressed, using NOTIFY_STREAM

    WARN: does not inform the message callback
    """
    if authorization is None:
        authorization = await state.config.authorize_notify(
            tracing=b"",  # TODO: traicng
            topic=message.topic,
            identifier=message.identifier,
            message_sha512=message.sha512,
            now=time.time(),
        )

    await send_notify_stream_given_first_headers(
        state=state,
        stream=message.stream,
        length=message.length,
        identifier=message.identifier,
        topic=message.topic,
        sha512=message.sha512,
        first_headers=serialize_s2b_notify_stream(
            S2B_NotifyStreamStartUncompressed(
                type=SubscriberToBroadcasterStatefulMessageType.NOTIFY_STREAM,
                authorization=authorization,
                tracing=b"",  # TODO: tracing
                identifier=message.identifier,
                part_id=None,
                topic=message.topic,
                compressor_id=None,
                uncompressed_length=message.length,
                unverified_uncompressed_sha512=message.sha512,
                payload=b"",
            ),
            minimal_headers=state.config.websocket_minimal_headers,
        ),
        msg=message,
    )


async def send_internal_small_message_compressed(
    state: StateOpen, message: InternalSmallMessage, /, *, compressor: CompressorReady
) -> None:
    """Sends the given internal small message compressed; will use NOTIFY
    if possible, otherwise NOTIFY_STREAM

    WARN: does not inform the message callback
    """
    with reserve_compressor(state, compressor) as reservation:
        compressed_data = reservation.compress(message.data)

    await send_internal_small_message_compressed_with_compressed_data(
        state,
        message,
        compressor_id=compressor.identifier,
        compressed_data=compressed_data,
    )


async def send_internal_large_message_compressed(
    state: StateOpen,
    message: InternalLargeMessage,
    /,
    *,
    compressor: CompressorReady,
) -> None:
    """Sends the given internal large message compressed; will use NOTIFY if
    possible, otherwise NOTIFY_STREAM

    WARN: does not inform the message callback
    """
    spool_size = state.config.max_websocket_message_size or (2**64 - 1)

    with tempfile.SpooledTemporaryFile(max_size=spool_size) as compressed:
        with reserve_compressor(state, compressor) as reservation:
            chunker = reservation.chunker(
                size=message.length, chunk_size=io.DEFAULT_BUFFER_SIZE
            )

            uncompressed_pos = 0
            compressed_pos = 0
            compressed_hasher = hashlib.sha512()
            while True:
                uncompressed_chunk = read_exact(
                    message.stream,
                    min(io.DEFAULT_BUFFER_SIZE, message.length - uncompressed_pos),
                )
                uncompressed_pos += len(uncompressed_chunk)

                for compressed_chunk in chunker.compress(uncompressed_chunk):
                    compressed.write(compressed_chunk)
                    compressed_pos += len(compressed_chunk)
                    compressed_hasher.update(compressed_chunk)

                if uncompressed_pos >= message.length:
                    break

                await asyncio.sleep(0)

            assert uncompressed_pos == message.length, "read_exact failure"
            for compressed_chunk in chunker.finish():
                compressed.write(compressed_chunk)
                compressed_pos += len(compressed_chunk)
                compressed_hasher.update(compressed_chunk)

        compressed_sha512 = compressed_hasher.digest()
        compressed_length = compressed_pos

        compressed.seek(0, os.SEEK_SET)
        if compressed_length > spool_size:
            authorization = await state.config.authorize_notify(
                tracing=b"",  # TODO: tracing
                topic=message.topic,
                identifier=message.identifier,
                message_sha512=compressed_sha512,
                now=time.time(),
            )
            return await send_notify_stream_given_first_headers(
                state=state,
                stream=compressed,
                length=compressed_length,
                identifier=message.identifier,
                topic=message.topic,
                sha512=compressed_sha512,
                first_headers=serialize_s2b_notify_stream(
                    S2B_NotifyStreamStartCompressed(
                        type=SubscriberToBroadcasterStatefulMessageType.NOTIFY_STREAM,
                        authorization=authorization,
                        tracing=b"",  # TODO: tracing
                        identifier=message.identifier,
                        part_id=None,
                        topic=message.topic,
                        compressor_id=compressor.identifier,
                        compressed_length=compressed_length,
                        decompressed_length=message.length,
                        unverified_compressed_sha512=compressed_sha512,
                        payload=b"",
                    ),
                    minimal_headers=state.config.websocket_minimal_headers,
                ),
                msg=message,
            )

        compressed_data = read_exact(compressed, compressed_length)

    await send_internal_small_message_compressed_with_compressed_data(
        state,
        message,
        compressor_id=compressor.identifier,
        compressed_data=compressed_data,
        compressed_sha512=compressed_sha512,
    )


async def send_internal_small_message_compressed_with_compressed_data(
    state: StateOpen,
    message: Union[InternalSmallMessage, InternalLargeMessage],
    /,
    *,
    compressor_id: int,
    compressed_data: bytes,
    compressed_sha512: Optional[bytes] = None,
) -> None:
    """Sends the given internal message where the uncompressed data may or may
    not be small enough to hold in memory but the compressed data is small enough
    to hold in memory. Uses NOTIFY if possible, otherwise NOTIFY_STREAM

    WARN: does not inform the message callback
    """
    if compressed_sha512 is None:
        compressed_sha512 = hashlib.sha512(compressed_data).digest()

    identifier = secrets.token_bytes(4)
    tracing = b""  # TODO: tracing
    authorization = await state.config.authorize_notify(
        tracing=tracing,
        topic=message.topic,
        identifier=identifier,
        message_sha512=compressed_sha512,
        now=time.time(),
    )

    message_length = (
        len(message.data)
        if message.type == InternalMessageType.SMALL
        else message.length
    )

    if (
        state.config.max_websocket_message_size is None
        or message_length + 128 < state.config.max_websocket_message_size
    ):
        as_simple_message = serialize_s2b_notify(
            S2B_NotifyCompressed(
                type=SubscriberToBroadcasterStatefulMessageType.NOTIFY,
                authorization=authorization,
                tracing=b"",  # TODO: tracing
                identifier=identifier,
                compressor_id=compressor_id,
                topic=message.topic,
                verified_compressed_sha512=compressed_sha512,
                compressed_message=compressed_data,
                decompressed_length=message_length,
            ),
            minimal_headers=state.config.websocket_minimal_headers,
        )
        if (
            state.config.max_websocket_message_size is None
            or len(as_simple_message) < state.config.max_websocket_message_size
        ):
            state.expected_acks.append(
                B2S_ConfirmNotify(
                    type=BroadcasterToSubscriberStatefulMessageType.CONFIRM_NOTIFY,
                    identifier=identifier,
                    subscribers=-1,
                    authorization=None,
                    tracing=b"",
                )
            )
            state.sent_notifications.append(message)
            await state.websocket.send_bytes(as_simple_message)
            return

    await send_notify_stream_given_first_headers(
        state=state,
        stream=io.BytesIO(compressed_data),
        length=len(compressed_data),
        identifier=identifier,
        topic=message.topic,
        sha512=message.sha512,
        first_headers=serialize_s2b_notify_stream(
            S2B_NotifyStreamStartCompressed(
                type=SubscriberToBroadcasterStatefulMessageType.NOTIFY_STREAM,
                authorization=authorization,
                tracing=b"",  # TODO: tracing
                identifier=identifier,
                part_id=None,
                topic=message.topic,
                compressor_id=compressor_id,
                compressed_length=len(compressed_data),
                decompressed_length=message_length,
                unverified_compressed_sha512=compressed_sha512,
                payload=b"",
            ),
            minimal_headers=state.config.websocket_minimal_headers,
        ),
        msg=message,
    )


async def send_notify_stream_given_first_headers(
    state: StateOpen,
    stream: SyncReadableBytesIO,
    length: int,
    identifier: bytes,
    topic: bytes,
    sha512: bytes,
    first_headers: bytes,
    msg: InternalMessage,
) -> None:
    """Sends the given stream of data to the broadcaster via potentiially multiple
    NOTIFY_STREAM messages, using the given headers for the first message

    WARN: does not inform the message callback

    Args:
        state (StateOpen): the state, for grabbing configuration options
        stream (SyncReadableBytesIO): the stream of data to send
        length (int): the length of the stream
        identifier (bytes): the arbitrary unique identifier assigned by the subscriber
            so the broadcaster can combine the related messages
        topic (bytes): the topic of the message
        sha512 (bytes): the sha512 of the data being sent (so if the stream is
            compressed, the sha512 of the compressed data)
        first_headers (bytes): the headers for the first message which lets the broadcaster
            know how to interpret the stream
        msg (InternalMessage): the message to add to sent_notifications before sending the
            first message
    """
    headers = first_headers
    msg_size = state.config.max_websocket_message_size or (2**64 - 1)

    part_id = 0
    pos = 0
    while True:
        target_read_amount = min(length - pos, max(512, msg_size - len(headers)))
        payload = (
            b"" if target_read_amount == 0 else read_exact(stream, target_read_amount)
        )
        pos += len(payload)
        is_done = pos >= length
        state.expected_acks.append(
            B2S_ConfirmNotify(
                type=BroadcasterToSubscriberStatefulMessageType.CONFIRM_NOTIFY,
                identifier=identifier,
                subscribers=-1,
                authorization=None,
                tracing=b"",
            )
            if is_done
            else B2S_ContinueNotify(
                type=BroadcasterToSubscriberStatefulMessageType.CONTINUE_NOTIFY,
                identifier=identifier,
                part_id=part_id,
                authorization=None,
                tracing=b"",
            )
        )
        if part_id == 0:
            state.sent_notifications.append(msg)
        await state.websocket.send_bytes(headers + payload)

        if is_done:
            return

        part_id += 1
        authorization = await state.config.authorize_notify(
            tracing=b"",
            topic=topic,
            identifier=identifier,
            message_sha512=sha512,
            now=time.time(),
        )
        headers = serialize_s2b_notify_stream(
            S2B_NotifyStreamContinuation(
                type=SubscriberToBroadcasterStatefulMessageType.NOTIFY_STREAM,
                authorization=authorization,
                tracing=b"",  # TODO: tracing
                identifier=identifier,
                part_id=part_id,
                payload=b"",
            ),
            minimal_headers=state.config.websocket_minimal_headers,
        )



import asyncio
from enum import Enum, auto
from typing import TYPE_CHECKING, Dict, List, Literal, Optional, Protocol, Union

from lonelypsp.compat import fast_dataclass

from lonelypsc.util.errors import combine_multiple_exceptions
from lonelypsc.ws.check_result import CheckResult

try:
    import zstandard
except ImportError:
    ...


class CompressorState(Enum):
    """Describes the state that a zstandard compressor is in"""

    PREPARING = auto()
    """We are in the process of preparing the compressor for use"""

    READY = auto()
    """The compressor is ready to use right now; the individual zstandard compressors/
    decompressors are not async or thread safe, but the pool of them is async safe
    """


@fast_dataclass
class CompressorReady:
    """A compressor which is ready to use"""

    type: Literal[CompressorState.READY]
    """discriminator value"""

    identifier: int
    """the integer identifier for this compressor; positive integers only."""

    level: int
    """the compression level the compressor is set to"""

    min_size: int
    """the minimum size, in bytes, inclusive, that a message can be for the subscriber
    to choose this compressor for the message. 0 means no minimum size. Note that the
    broadcaster may use this compressor for messages smaller than this size and the
    subscriber will still decompress it.
    """

    max_size: Optional[int]
    """the maximum size, if any, in bytes, exclusive, that a message can be for the subscriber
    to choose this compressor for the message. None means no maximum size. Note that the
    broadcaster may use this compressor for messages larger than this size and the
    subscriber will still decompress it.
    """

    data: "Optional[zstandard.ZstdCompressionDict]"
    """if there is a custom compression dictionary, that dictionary, otherwise None"""

    compressors: "List[zstandard.ZstdCompressor]"
    """the zstandard compressor objects that are not in use. pulled LIFO as it is
    preferable to reuse the same object as much as possible.
    
    WARN: individual compressors are not asyncio safe
    """

    decompressors: "List[zstandard.ZstdDecompressor]"
    """the zstandard decompressors that are not in use. pulled LIFO as it is
    preferable to reuse the same object as much as possible.
    
    WARN: decompressors are not asyncio-safe
    """


@fast_dataclass
class CompressorPreparing:
    """A compressor that isn't ready to use yet"""

    type: Literal[CompressorState.PREPARING]
    """discriminator value"""

    identifier: int
    """the integer identifier for this compressor; positive integers only.

    1 is reserved for the compressor with no special compression dictionary,
    i.e., data is compressed by building a compression dictionary for each
    message. this works best on moderate sized messages (e.g., at least 16kb)
    as otherwise the overhead of the dictionary size may outweigh the benefits
    of using it

    2-65535 are reserved for preset compression dictionaries, where the compression
    dictionary is distributed out of band

    65536 and above are custom dictionaries built for the connection
    """

    task: asyncio.Task[CompressorReady]
    """The task that is working on preparing the compressor for use"""


Compressor = Union[CompressorReady, CompressorPreparing]


class CompressorStore(Protocol):
    """Describes something that can store compressors and retrieve them
    either by identifier or find the appropriate one for a given message
    length. Also needs to be able to sweep preparing compressors and check
    if any are ready
    """

    def add_compressor(self, compressor: Compressor) -> None:
        """Add a compressor to the store.

        Raises ValueError if the identifier is already in the store
        """

    def remove_compressor(self, identifier: int) -> None:
        """Remove a compressor from the store.

        Raises KeyError if the identifier is not in the store
        """

    def get_for_compression(self, length: int) -> Optional[Compressor]:
        """Get the compressor to use for a message of the given length, or
        None to not use compression
        """

    def get_for_decompression(self, identifier: int) -> Compressor:
        """Get the compressor to use to decompress a message compressed
        with the given identifier

        Raises KeyError if the identifier is not in the store
        """

    def get_compressor_tasks(self) -> List[asyncio.Task[CompressorReady]]:
        """Get the tasks that are preparing compressors, or an empty list
        if there are none.
        """

    def check_compressor_tasks(self) -> CheckResult:
        """Checks all tasks to see if any are done. For finished tasks,
        replace the compressor internally with the ready compressor.

        Raises any errors that occurred in the tasks

        Returns `CheckResult.RESTART` if any tasks were done, otherwise
        `CheckResult.CONTINUE`
        """


class CompressorStoreImpl:
    """Satisfies CompressorStore using separate lists; intended for not that many
    compressors and using them is much more common than mutating them

    This is the only expected implementation of CompressorStore right now, and
    normally it's unhelpful to make a protocol if there is only one
    implementation and its not easy for the caller to swap it out, but in this
    case it feels better than exposing the data structure chosen here in the
    type hints (even if underscored), since it might be an important part of the
    hot path (i.e., might need to change it later for performance reasons) while
    at the same time should not be particularly usecase dependent (so letting
    callers swap the implementation just bloats the API)
    """

    def __init__(self) -> None:
        self.preparing: List[CompressorPreparing] = []
        """Compressors that are not ready to use yet, in insertion order"""

        self.ready: List[CompressorReady] = []
        """Compressors that are ready to use, sorted by ascending minimum size"""

        self.by_compressor_id: Dict[int, Compressor] = {}
        """Lookup table for compressors by identifier"""

    def add_compressor(self, compressor: Compressor) -> None:
        if compressor.identifier in self.by_compressor_id:
            raise ValueError(
                f"compressor with identifier {compressor.identifier} already exists"
            )

        if compressor.type == CompressorState.PREPARING:
            self.preparing.append(compressor)
        else:
            self._insert_ready(compressor)

        self.by_compressor_id[compressor.identifier] = compressor

    def remove_compressor(self, identifier: int) -> None:
        compressor = self.by_compressor_id.pop(identifier)
        if compressor.type == CompressorState.PREPARING:
            self.preparing.remove(compressor)
        else:
            self.ready.remove(compressor)

    def get_for_compression(self, length: int) -> Optional[Compressor]:
        for compressor in self.ready:
            if compressor.min_size <= length < (compressor.max_size or float("+inf")):
                return compressor

        return None

    def get_for_decompression(self, identifier: int) -> Compressor:
        return self.by_compressor_id[identifier]

    def get_compressor_tasks(self) -> List[asyncio.Task[CompressorReady]]:
        return [compressor.task for compressor in self.preparing]

    def check_compressor_tasks(self) -> CheckResult:
        if not self.preparing:
            return CheckResult.CONTINUE

        new_preparing: List[CompressorPreparing] = []
        excs: List[BaseException] = []
        for compressor in self.preparing:
            if not compressor.task.done():
                new_preparing.append(compressor)
                continue

            try:
                ready = compressor.task.result()
                if excs:
                    continue
                self._insert_ready(ready)
                self.by_compressor_id[ready.identifier] = ready
            except BaseException as e:
                excs.append(e)

        if excs:
            raise combine_multiple_exceptions("failed to prepare compressors", excs)

        if len(new_preparing) == len(self.preparing):
            return CheckResult.CONTINUE

        self.preparing = new_preparing
        return CheckResult.RESTART

    def _insert_ready(self, compressor: CompressorReady) -> None:
        for i, existing in enumerate(self.ready):
            if existing.min_size > compressor.min_size:
                self.ready.insert(i, compressor)
                break
        else:
            self.ready.append(compressor)


if TYPE_CHECKING:
    _: CompressorStore = CompressorStoreImpl()



import asyncio
import dataclasses
import hashlib
import tempfile
import time
from typing import TYPE_CHECKING

from lonelypsp.auth.config import AuthResult
from lonelypsp.stateful.constants import SubscriberToBroadcasterStatefulMessageType
from lonelypsp.stateful.messages.receive_stream import B2S_ReceiveStream

from lonelypsc.client import PubSubError, PubSubIrrecoverableError
from lonelypsc.ws.handlers.open.handle_authorized_receive import (
    handle_authorized_receive,
)
from lonelypsc.ws.handlers.open.messages.protocol import MessageChecker
from lonelypsc.ws.handlers.open.websocket_url import (
    make_for_receive_websocket_url_and_change_counter,
)
from lonelypsc.ws.state import (
    ReceivingAuthorizing,
    ReceivingIncomplete,
    ReceivingState,
    StateOpen,
    UnsentAckConfirmReceive,
    UnsentAckContinueReceive,
)


def check_receive_stream(state: StateOpen, message: B2S_ReceiveStream) -> None:
    """Handles the subscriber receiving the a message from the broadcaster that
    a message was sent to a specific topic that the subscriber is subscribed to;
    if the message is long, this may only contain a part of the message.

    Raises an exception if this is not a proper continuation (either a new message
    after the last one finished normally, or the continuation from the last message)
    """
    spool_size = (
        state.config.max_websocket_message_size
        if state.config.max_websocket_message_size is not None
        else 2**64 - 1
    )
    if message.part_id is None:
        if state.receiving is not None:
            if state.receiving.type != ReceivingState.INCOMPLETE:
                raise PubSubIrrecoverableError(
                    "invariant violated: check_receive_stream while not in INCOMPLETE state"
                )

            raise PubSubError(
                f"expected continnuation of {state.receiving.first.identifier!r}, got new message {message.identifier!r}"
            )

        state.receiving = ReceivingIncomplete(
            type=ReceivingState.INCOMPLETE,
            first=dataclasses.replace(message, payload=b""),
            part_id=-1,
            body_hasher=hashlib.sha512(),
            body=tempfile.SpooledTemporaryFile(max_size=spool_size),
            authorization_task=asyncio.create_task(
                state.config.is_receive_allowed(
                    tracing=message.tracing,
                    url=make_for_receive_websocket_url_and_change_counter(state),
                    topic=message.topic,
                    identifier=message.identifier,
                    message_sha512=(
                        message.unverified_compressed_sha512
                        if message.compressor_id is not None
                        else message.unverified_uncompressed_sha512
                    ),
                    now=time.time(),
                    authorization=message.authorization,
                )
            ),
        )

    if state.receiving is None:
        raise PubSubError(
            f"expected new message, got continuation of {message.identifier!r}"
        )

    if state.receiving.type != ReceivingState.INCOMPLETE:
        raise PubSubIrrecoverableError(
            "invariant violated: check_receive_stream while not in INCOMPLETE state"
        )

    if message.identifier != state.receiving.first.identifier:
        raise PubSubError(
            f"expected continuation of {state.receiving.first.identifier!r}, got {message.identifier!r}"
        )

    if state.receiving.part_id + 1 != (message.part_id or 0):
        raise PubSubError(
            f"expected part {state.receiving.part_id + 1}, got {message.part_id!r}"
        )

    if (
        state.receiving.authorization_task is not None
        and state.receiving.authorization_task.done()
    ):
        result = state.receiving.authorization_task.result()
        if result != AuthResult.OK:
            raise PubSubError(f"authorization failed: {result}")
        state.receiving.authorization_task = None

    state.receiving.body.write(message.payload)
    state.receiving.body_hasher.update(message.payload)
    state.receiving.part_id += 1

    received_length = state.receiving.body.tell()
    expected_length = (
        state.receiving.first.compressed_length
        if state.receiving.first.compressor_id is not None
        else state.receiving.first.uncompressed_length
    )

    if received_length > expected_length:
        raise PubSubError(
            f"received more data than expected: {received_length} > {expected_length}"
        )

    if received_length < expected_length:
        state.unsent_acks.append(
            UnsentAckContinueReceive(
                type=SubscriberToBroadcasterStatefulMessageType.CONTINUE_RECEIVE,
                identifier=state.receiving.first.identifier,
                part_id=state.receiving.part_id,
            )
        )
        return

    received_hash = state.receiving.body_hasher.digest()
    expected_hash = (
        state.receiving.first.unverified_compressed_sha512
        if state.receiving.first.compressor_id is not None
        else state.receiving.first.unverified_uncompressed_sha512
    )

    if received_hash != expected_hash:
        raise PubSubError(
            f"received data hash {received_hash.hex()} does not match expected {expected_hash.hex()}"
        )

    state.unsent_acks.append(
        UnsentAckConfirmReceive(
            type=SubscriberToBroadcasterStatefulMessageType.CONFIRM_RECEIVE,
            identifier=state.receiving.first.identifier,
        )
    )

    if state.receiving.authorization_task is not None:
        state.receiving = ReceivingAuthorizing(
            type=ReceivingState.AUTHORIZING,
            first=state.receiving.first,
            body=state.receiving.body,
            authorization_task=state.receiving.authorization_task,
        )
        return

    handle_authorized_receive(state, state.receiving.first, state.receiving.body)


if TYPE_CHECKING:
    _: MessageChecker[B2S_ReceiveStream] = check_receive_stream



import asyncio
import hashlib
import io
import os
import tempfile
from typing import IO, cast

from lonelypsp.stateful.messages.receive_stream import B2S_ReceiveStreamStartCompressed
from lonelypsp.stateful.parser_helpers import read_exact

from lonelypsc.client import PubSubError
from lonelypsc.types.sync_io import SyncIOBaseLikeIO
from lonelypsc.ws.check_result import CheckResult
from lonelypsc.ws.compressor import CompressorReady, CompressorState
from lonelypsc.ws.handlers.open.compressor_utils import (
    DecompressorReservation,
)
from lonelypsc.ws.state import (
    ReceivedLargeMessage,
    ReceivedMessage,
    ReceivedMessageType,
    ReceivedSmallMessage,
    ReceivingDecompressing,
    ReceivingState,
    StateOpen,
)


def _decompress(
    compressed_body: SyncIOBaseLikeIO,
    first: B2S_ReceiveStreamStartCompressed,
    compressor: CompressorReady,
    spool_size: int,
    decompression_max_window_size: int,
) -> ReceivedMessage:
    try:
        decompressed = tempfile.SpooledTemporaryFile(max_size=spool_size)
        try:
            with DecompressorReservation(
                compressor,
                max_window_size=decompression_max_window_size,
                max_decompressors=5,
            ) as decompressor, decompressor.stream_reader(
                cast(IO[bytes], compressed_body)
            ) as reader:
                read_so_far = 0
                hasher = hashlib.sha512()
                while True:
                    chunk = reader.read(io.DEFAULT_BUFFER_SIZE)
                    if not chunk:
                        break

                    if read_so_far + len(chunk) > first.decompressed_length:
                        raise PubSubError("decompressed data is longer than expected")

                    decompressed.write(chunk)
                    read_so_far += len(chunk)
                    hasher.update(chunk)

                if read_so_far != first.decompressed_length:
                    raise PubSubError("decompressed data is shorter than expected")

                decompressed_sha512 = hasher.digest()

            compressed_body.close()
            decompressed.seek(0, os.SEEK_SET)
            decompressed_msg: ReceivedMessage
            if read_so_far < spool_size:
                decompressed_msg = ReceivedSmallMessage(
                    type=ReceivedMessageType.SMALL,
                    topic=first.topic,
                    data=read_exact(decompressed, first.decompressed_length),
                    sha512=decompressed_sha512,
                )
                decompressed.close()
            else:
                decompressed_msg = ReceivedLargeMessage(
                    type=ReceivedMessageType.LARGE,
                    topic=first.topic,
                    stream=decompressed,
                    sha512=decompressed_sha512,
                )

            return decompressed_msg
        except BaseException:
            decompressed.close()
            raise
    finally:
        compressed_body.close()


def check_receiving_waiting_compressor(state: StateOpen) -> CheckResult:
    """
    Tries to move receiving notify state `WAITING_COMPRESSOR` to `DECOMPRESSING`

    Raises an error if it is not possible to decompress the message; since the
    subscriber is not checking the read_task, it won't parse `DISABLE_ZSTD_CUSTOM`
    that were sent after the `RECEIVE_STREAM`, so if its not decompressable it was
    not decompressible when it was received (i.e., that error is not a race condition)
    """
    if (
        state.receiving is None
        or state.receiving.type != ReceivingState.WAITING_COMPRESSOR
    ):
        return CheckResult.CONTINUE

    try:
        compressor = state.compressors.get_for_decompression(
            state.receiving.first.compressor_id
        )
    except KeyError:
        raise PubSubError(f"unknown compressor {state.receiving.first.compressor_id}")

    if compressor.type == CompressorState.PREPARING:
        return CheckResult.CONTINUE

    state.receiving = ReceivingDecompressing(
        type=ReceivingState.DECOMPRESSING,
        task=asyncio.create_task(
            asyncio.to_thread(
                _decompress,
                state.receiving.compressed_body,
                state.receiving.first,
                compressor,
                state.config.max_websocket_message_size or (2**64 - 1),
                state.config.decompression_max_window_size,
            )
        ),
    )
    return CheckResult.RESTART



from typing import TYPE_CHECKING

from lonelypsp.stateful.messages.confirm_subscribe import (
    B2S_ConfirmSubscribeExact,
)

from lonelypsc.client import PubSubError
from lonelypsc.ws.handlers.open.messages.protocol import MessageChecker
from lonelypsc.ws.state import StateOpen


def check_confirm_subscribe_exact(
    state: StateOpen, message: B2S_ConfirmSubscribeExact
) -> None:
    """Handles the subscriber receiving a message from the broadcaster that it has
    successfully subscribed to a specific topic, raising an error if this isn't the
    acknowledgement the subscriber was expecting.
    """
    try:
        expected = state.expected_acks.popleft()
    except IndexError:
        raise PubSubError(f"{message.type} received when not expecting any acks")

    if expected.type != message.type:
        raise PubSubError(f"expected {expected.type}, got {message.type}")
    if expected.topic != message.topic:
        raise PubSubError(
            f"expected {message.type} {expected.topic!r}, got {message.topic!r}"
        )

    state.exact_subscriptions.add(expected.topic)


if TYPE_CHECKING:
    _: MessageChecker[B2S_ConfirmSubscribeExact] = check_confirm_subscribe_exact



from typing import Dict, cast

from lonelypsp.stateful.constants import BroadcasterToSubscriberStatefulMessageType
from lonelypsp.stateful.message import B2S_Message

from lonelypsc.client import PubSubError
from lonelypsc.ws.handlers.open.messages.confirm_notify import check_confirm_notify
from lonelypsc.ws.handlers.open.messages.confirm_subscribe_exact import (
    check_confirm_subscribe_exact,
)
from lonelypsc.ws.handlers.open.messages.confirm_subscribe_glob import (
    check_confirm_subscribe_glob,
)
from lonelypsc.ws.handlers.open.messages.confirm_unsubscribe_exact import (
    check_confirm_unsubscribe_exact,
)
from lonelypsc.ws.handlers.open.messages.confirm_unsubscribe_glob import (
    check_confirm_unsubscribe_glob,
)
from lonelypsc.ws.handlers.open.messages.continue_notify import check_continue_notify
from lonelypsc.ws.handlers.open.messages.disable_zstd_custom import (
    check_disable_zstd_custom,
)
from lonelypsc.ws.handlers.open.messages.enable_zstd_custom import (
    check_enable_zstd_custom,
)
from lonelypsc.ws.handlers.open.messages.enable_zstd_preset import (
    check_enable_zstd_preset,
)
from lonelypsc.ws.handlers.open.messages.missed import check_missed
from lonelypsc.ws.handlers.open.messages.protocol import MessageChecker
from lonelypsc.ws.handlers.open.messages.receive_stream import check_receive_stream
from lonelypsc.ws.state import StateOpen

_handlers = cast(
    Dict[
        BroadcasterToSubscriberStatefulMessageType,
        MessageChecker[B2S_Message],  # constrained further by the key
    ],
    {
        BroadcasterToSubscriberStatefulMessageType.CONFIRM_NOTIFY: check_confirm_notify,
        BroadcasterToSubscriberStatefulMessageType.CONFIRM_SUBSCRIBE_EXACT: check_confirm_subscribe_exact,
        BroadcasterToSubscriberStatefulMessageType.CONFIRM_SUBSCRIBE_GLOB: check_confirm_subscribe_glob,
        BroadcasterToSubscriberStatefulMessageType.CONFIRM_UNSUBSCRIBE_EXACT: check_confirm_unsubscribe_exact,
        BroadcasterToSubscriberStatefulMessageType.CONFIRM_UNSUBSCRIBE_GLOB: check_confirm_unsubscribe_glob,
        BroadcasterToSubscriberStatefulMessageType.CONTINUE_NOTIFY: check_continue_notify,
        BroadcasterToSubscriberStatefulMessageType.DISABLE_ZSTD_CUSTOM: check_disable_zstd_custom,
        BroadcasterToSubscriberStatefulMessageType.ENABLE_ZSTD_CUSTOM: check_enable_zstd_custom,
        BroadcasterToSubscriberStatefulMessageType.ENABLE_ZSTD_PRESET: check_enable_zstd_preset,
        BroadcasterToSubscriberStatefulMessageType.MISSED: check_missed,
        BroadcasterToSubscriberStatefulMessageType.RECEIVE_STREAM: check_receive_stream,
    },
)


def check_read_message(state: StateOpen, message: B2S_Message) -> None:
    """Handles receiving the given message from the broadcaster in the given state,
    raising an exception if it does not make sense

    Does not support RECEIVE_STREAM which has a slightly different function signature
    """
    handler = _handlers.get(message.type)
    if handler is None:
        raise PubSubError(f"unexpected message type: {message.type}")
    handler(state, message)



import asyncio
import math
import sys
import time
from typing import Any, List

from lonelypsp.util.cancel_and_check import cancel_and_check

from lonelypsc.ws.state import (
    OpenRetryInformationType,
    ReceivingState,
    SendingState,
    StateOpen,
)

if sys.version_info >= (3, 11):
    from typing import Never
else:
    from typing import NoReturn as Never


async def wait_something_changed(state: StateOpen) -> None:
    """Waits for something to do"""
    managed_by_us: List[asyncio.Task[Any]] = []
    other_tasks: List[asyncio.Task[Any]] = []

    try:
        managed_by_us.append(asyncio.create_task(state.cancel_requested.wait()))
        if state.retry.type == OpenRetryInformationType.TENTATIVE:
            managed_by_us.append(
                asyncio.create_task(
                    asyncio.sleep(math.ceil(state.retry.stable_at - time.time()))
                )
            )

        other_tasks.extend(state.compressors.get_compressor_tasks())

        if state.unsent_notifications.empty():
            managed_by_us.append(
                asyncio.create_task(state.unsent_notifications.wait_not_empty())
            )

        other_tasks.extend(
            t.callback.task
            for t in state.resending_notifications
            if t.callback.task is not None
        )
        other_tasks.extend(
            t.callback.task
            for t in state.sent_notifications
            if t.callback.task is not None
        )

        if state.management_tasks.empty():
            managed_by_us.append(
                asyncio.create_task(state.management_tasks.wait_not_empty())
            )

        if state.receiving is None:
            ...  # avoids a nesting level
        elif state.receiving.type == ReceivingState.INCOMPLETE:
            ...  # waiting on read_task
        elif (
            state.receiving.type == ReceivingState.AUTHORIZING
            or state.receiving.type == ReceivingState.AUTHORIZING_MISSED
        ):
            other_tasks.append(state.receiving.authorization_task)
        elif state.receiving.type == ReceivingState.WAITING_COMPRESSOR:
            ...  # already have compressor tasks in other_tasks
        elif state.receiving.type == ReceivingState.DECOMPRESSING:
            other_tasks.append(state.receiving.task)
        else:
            _assert_never(state.receiving)

        if state.received.full():
            managed_by_us.append(asyncio.create_task(state.received.wait_not_full()))

        if state.sending is None:
            ...  # avoids a nesting level
        elif state.sending.type == SendingState.SIMPLE:
            other_tasks.append(state.sending.task)
        elif state.sending.type == SendingState.MANAGEMENT_TASK:
            other_tasks.append(state.sending.task)
        elif state.sending.type == SendingState.INTERNAL_MESSAGE:
            if state.sending.internal_message.callback.task is not None:
                other_tasks.append(state.sending.internal_message.callback.task)
            other_tasks.append(state.sending.task)
        else:
            _assert_never(state.sending)

        other_tasks.append(state.read_task)
        other_tasks.extend(state.backgrounded)

        await asyncio.wait(
            managed_by_us + other_tasks, return_when=asyncio.FIRST_COMPLETED
        )
    finally:
        await asyncio.gather(*(cancel_and_check(task) for task in managed_by_us))


def _assert_never(v: Never) -> None:
    raise AssertionError(f"Unhandled value: {v!r}")



import asyncio

from aiohttp import ClientWebSocketResponse, WSMsgType

from lonelypsc.types.websocket_message import WSMessage


def make_websocket_read_task(
    websocket: ClientWebSocketResponse,
) -> asyncio.Task[WSMessage]:
    """Creats an asyncio task that provides a better typed version of websocket.receive()"""
    return asyncio.create_task(adapt_websocket_read(websocket))


async def adapt_websocket_read(websocket: ClientWebSocketResponse) -> WSMessage:
    """Adapts the return type of websocket.receive() to WSMessage (the fastapi type)"""

    result = await websocket.receive()
    if result.type == WSMsgType.BINARY:
        return {"type": "websocket.receive", "bytes": result.data}

    if result.type == WSMsgType.CLOSE or result.type == WSMsgType.CLOSED:
        return {
            "type": "websocket.disconnect",
            "code": 1000,
            "reason": result.extra or "",
        }

    if result.type == WSMsgType.ERROR:
        return {
            "type": "websocket.disconnect",
            "code": 1011,
            "reason": result.extra or "",
        }

    if result.type == WSMsgType.TEXT:
        return {"type": "websocket.receive", "text": result.data}

    raise ValueError(f"Unexpected WSMsgType: {result}")



import asyncio
import hashlib
import os
import re
from dataclasses import dataclass
from enum import Enum, auto
from io import BytesIO
from types import TracebackType
from typing import (
    TYPE_CHECKING,
    Awaitable,
    Callable,
    ContextManager,
    Dict,
    Generic,
    Iterable,
    List,
    Literal,
    Optional,
    Protocol,
    Set,
    Type,
    TypeVar,
    Union,
    overload,
)

from lonelypsp.compat import fast_dataclass
from lonelypsp.stateless.make_strong_etag import StrongEtag
from lonelypsp.util.cancel_and_check import cancel_and_check
from lonelypsp.util.drainable_asyncio_queue import DrainableAsyncioQueue

from lonelypsc.types.sync_io import (
    SyncReadableBytesIO,
    SyncStandardIO,
)
from lonelypsc.util.errors import combine_multiple_exceptions
from lonelypsc.util.io_helpers import PositionedSyncStandardIO

try:
    from glob import translate as _glob_translate  # type: ignore

    def translate(pat: str) -> str:
        return _glob_translate(pat, recursive=True, include_hidden=True)

except ImportError:
    from fnmatch import translate


class _Cleanup(Protocol):
    async def __call__(self) -> None: ...


@dataclass
class PubSubClientMessageWithCleanup:
    """A message received on a topic plus the function to cleanup any associated
    resources
    """

    topic: bytes
    """The topic the message was sent to"""
    sha512: bytes
    """The sha512 hash of the message"""
    data: SyncReadableBytesIO
    """The message data"""
    cleanup: _Cleanup
    """The function to cleanup the message once it's removed from the buffer and done with"""


class PubSubClientMessage(Protocol):
    """A message received on a topic where cleanup is already being handled"""

    @property
    def topic(self) -> bytes:
        """The topic the message was sent to"""

    @property
    def sha512(self) -> bytes:
        """The sha512 hash of the message"""

    @property
    def data(self) -> SyncReadableBytesIO:
        """The message data"""


if TYPE_CHECKING:
    _: Type[PubSubClientMessage] = PubSubClientMessageWithCleanup


class PubSubClientSubscriptionIterator:
    def __init__(
        self,
        state: "_PubSubClientSubscriptionStateEnteredBuffering",
    ) -> None:
        self.state = state

    def __aiter__(self) -> "PubSubClientSubscriptionIterator":
        return self

    async def __anext__(self) -> PubSubClientMessage:
        """Explicitly expects cancellation"""
        while True:
            try:
                item_to_cleanup = self.state.cleanup.get_nowait()
            except asyncio.QueueEmpty:
                break
            await item_to_cleanup.cleanup()

        seen_lost = self.state.status == PubSubClientConnectionStatus.LOST
        while True:
            if not self.state.status_queue.empty():
                new_status = self.state.status_queue.get_nowait()
                self.state.status = new_status
                if self.state.status == PubSubClientConnectionStatus.LOST:
                    seen_lost = True
                continue

            if self.state.status == PubSubClientConnectionStatus.LOST:
                await self.state.status_queue.wait_not_empty()
                continue

            if self.state.status == PubSubClientConnectionStatus.ABANDONED:
                raise PubSubRequestConnectionAbandonedError("connection abandoned")

            assert self.state.status == PubSubClientConnectionStatus.OK, "impossible"
            if seen_lost and self.state.on_receiving is not None:
                await self.state.on_receiving()
                seen_lost = False
                continue

            if self.state.buffer.empty():
                buffer_task = asyncio.create_task(self.state.buffer.wait_not_empty())
                status_task = asyncio.create_task(
                    self.state.status_queue.wait_not_empty()
                )
                try:
                    await asyncio.wait(
                        [buffer_task, status_task], return_when=asyncio.FIRST_COMPLETED
                    )
                finally:
                    await asyncio.gather(
                        cancel_and_check(buffer_task), cancel_and_check(status_task)
                    )
                continue

            result = self.state.buffer.get_nowait()
            try:
                self.state.cleanup.put_nowait(result)
            except asyncio.QueueFull:
                await result.cleanup()
                raise

            return result


class PubSubClientSubscriptionWithTimeoutIterator:
    """Wrapper around PubSubClientSubscriptionIterator that yields None if no message
    is received within the timeout (seconds). Note the timeout starts with `__anext__` is
    called, not when the last message was received.
    """

    def __init__(
        self,
        raw_iter: "PubSubClientSubscriptionIterator",
        timeout: float,
    ) -> None:
        self.raw_iter = raw_iter
        self.timeout = timeout

    def __aiter__(self) -> "PubSubClientSubscriptionWithTimeoutIterator":
        return self

    async def __anext__(self) -> Optional[PubSubClientMessage]:
        """Explicitly expects cancellation"""
        timeout_task = asyncio.create_task(asyncio.sleep(self.timeout))
        message_task = asyncio.create_task(self.raw_iter.__anext__())
        exc: Optional[BaseException] = None
        try:
            await asyncio.wait(
                (timeout_task, message_task), return_when=asyncio.FIRST_COMPLETED
            )
        except BaseException as e:
            exc = e

        _, message_result = await asyncio.gather(
            cancel_and_check(timeout_task, False), cancel_and_check(message_task)
        )

        if message_result is not None:
            return message_result

        if exc is not None:
            raise exc

        return None


_STATE_NOT_ENTERED: Literal[1] = 1
_STATE_ENTERED_NOT_BUFFERING: Literal[2] = 2
_STATE_ENTERED_BUFFERING: Literal[3] = 3
_STATE_DISPOSED: Literal[4] = 4


class OnReceiving(Protocol):
    async def __call__(self) -> None: ...


@dataclass
class _PubSubClientSubscriptionStateNotEntered:
    """State when the subscription has not yet been entered"""

    type: Literal[1]
    """Type discriminator (_STATE_NOT_ENTERED)"""

    client: "PubSubClient"
    """The client we are connected to"""

    exact: Set[bytes]
    """The exact topics that have been queued up to subscribe to when entered"""

    glob: Set[str]
    """The glob topics that have been queued up to subscribe to when entered"""

    on_receiving: Optional[OnReceiving]
    """The function to call at the beginning of what we believe to be a continuous
    stream of messages
    """


@dataclass
class _PubSubClientSubscriptionStateEnteredNotBuffering:
    """State when we have been entered but an iterator hasn't been created yet;
    we have subscribed to the topics but are not yet receiving/buffering messages
    """

    type: Literal[2]
    """Type discriminator (_STATE_ENTERED_NOT_BUFFERING)"""

    client: "PubSubClient"
    """The client we are connected to"""

    exact: Dict[bytes, int]
    """The topic -> subscription id pairs we are subscribed to"""

    glob: Dict[str, int]
    """The glob -> subscription id pairs we are subscribed to"""

    on_receiving: Optional[OnReceiving]
    """The function to call at the beginning of what we believe to be a continuous
    stream of messages
    """


@dataclass
class _PubSubClientSubscriptionStateEnteredBuffering:
    """State when we have been entered and an iterator has been created; we
    are subscribed and buffering messages
    """

    type: Literal[3]
    """Type discriminator (_STATE_ENTERED_BUFFERING)"""

    client: "PubSubClient"
    """The client we are connected to"""

    on_message_subscription_id: int
    """The registration id for the on_message callback"""

    on_status_subscription_id: int
    """The registration id for the connection status listeners"""

    status: "PubSubClientConnectionStatus"
    """The last connection status that we handled"""

    status_queue: DrainableAsyncioQueue["PubSubClientConnectionStatus"]
    """When the status changes we push the new status to this queue as we need
    to process them in order.
    """

    exact: Dict[bytes, int]
    """The topic -> subscription_id mapping for exact subscriptions. We care
    about a message because its an exact match if its a key in this dict.
    """

    glob_regexes: List[re.Pattern]
    """The list of regexes we are subscribed to, in the exact order of `glob_list`. We
    care about a message because its a glob match if any of these regexes match.
    """

    glob_list: List[str]
    """The globs that made the regexes in glob_regexes"""

    glob: Dict[str, int]
    """The glob -> subscription id pairs we are subscribed to"""

    buffer: DrainableAsyncioQueue[PubSubClientMessageWithCleanup]
    """the buffer that we push matching messages to such that they are read by anext"""

    cleanup: DrainableAsyncioQueue[PubSubClientMessageWithCleanup]
    """the messages that haven't been cleaned up yet but need to be; this is normally cleared
    out when calling anext on the iterator, but the last item has to be cleaned out when the
    subscription is exited
    """

    on_receiving: Optional[OnReceiving]
    """The function to call at the beginning of what we believe to be a continuous
    stream of messages
    """


@dataclass
class _PubSubClientSubscriptionStateDisposed:
    type: Literal[4]
    """Type discriminator (_STATE_DISPOSED)"""


_PubSubClientSubscriptionState = Union[
    _PubSubClientSubscriptionStateNotEntered,
    _PubSubClientSubscriptionStateEnteredNotBuffering,
    _PubSubClientSubscriptionStateEnteredBuffering,
    _PubSubClientSubscriptionStateDisposed,
]


class PubSubClientSubscription:
    """Describes a subscription to one or more topic/globs within a single
    client. When the client exits it will exit all subscriptions, but exiting
    a subscription does not exit the client.

    ## Usage

    Using the http client or you don't care about missed notifications:

    ```python
    async with client.subscribe_exact(b"topic") as subscription:
        async for message in await subscription.with_timeout(5):
            if message is None:
                print("still waiting")
            print(f"received message: {message.data.read()}")
    ```

    Using the websocket client and you care about missed notifications:

    ```python
    async def on_receiving():
        # reset your state here via polling. while you're doing this, any
        # messages received will be buffered and then played after.
        # usecases:
        #  - if incoming messages bust the cache, bust the cache here in case you
        #    missed a message to do so
        #  - if you are maintaining state, set the current state here (make sure to
        #    have some kind of counter in the log so you can discard messages already
        #    incorporated later)
        ...

    async with client.subscribe_exact(
        b"topic",
        on_receiving=on_receiving
    ) as subscription:
        async for message in await subscription.with_timeout(5):
            if message is None:
                print("still waiting")
            print(f"received message: {message.data.read()}")
    ```

    ## Details

    You can subscribe before aenter, which will queue up commands to be executed
    while aenter'ing. This facilitates generating methods for subscriptions that
    haven't been aenter'd yet.

    This interface is mostly to automatically pair subscribe/unsubscribes and, when
    creating an iterator, to handle buffering of messages between `anext` calls

    You can use this as an asynchronous iterable,, but you can only make one
    async iterator or you will get an error. This is because, since PEP 533 was
    not accepted, we cannot reliably detect when an iterator has been closed.
    However, there is a period of time while you are processing a message
    between `__anext__` calls where any message received needs to be buffered
    for the next `__anext__` call.

    While this is fine so long as the `__anext__` call is actually coming, if an
    iterator is leaked (i.e., no longer getting __anext__ calls) then the buffer
    will "quietly" grow.

    Thus, a safe way to use this is as follows:

    ```
    async with client.subscribe_exact(b"topic") as subscription: # subscribes
        async for message in await subscription.messages():  # starts buffering
            ...
    # exiting the subscription unsubscribes and stops buffering
    ```

    but this would be extremely error-prone if we allowed it:

    ```
    async with client.subscribe_exact(b"topic") as subscription:  # subscribes
        async for message in await subscription.messages():  # starts buffering
            break

        # still buffering! furthermore, lonelypsc doesn't
        # know if the other iter is still going or not! thus,
        # this is ambiguous!
        async for message in await subscription.messages():  # ERROR
            ...
    # exiting the subscription unsubscribes and stops buffering
    ```

    Note the buffering process does not start until the aiter is created, so if
    you want to only have the subscribe endpoint called once but create multiple
    iterators, and you are ok dropping messages between the iterators, this
    pattern will work:

    ```
    async with client.subscribe_exact(b"topic"):  # subscribes
        async with client.subscribe_exact(b"topic") as subscription: # no-op
            async for message in await subscription.messages(): # starts buffering
                ...
        # exiting the subscription stops buffering

        # any messages created before the next iterator is created are dropped

        async with client.subscribe_exact(b"topic") as subscription:  # no-op
            async for message in await subscription.messages():  # starts buffering
                ...
        # exiting the subscription stops buffering
    # exiting the last subscription to `topic` unsubscribes
    ```

    and if the potential dropped messages needs to be avoided but you cannot use
    a single `async for` loop, then don't use the `async for` syntax at all:
    instead use the `__anext__` method directly:

    ```
    async with client.subscribe_exact(b"topic") as subscription:  # subscribes
        my_iter = await subscription.messages()  # starts buffering

        # whenever you want...
        message = await my_iter.__anext__()  # gets a message, blocking if required
    # exiting the subscription unsubscribes and stops buffering
    ```
    """

    def __init__(
        self,
        client: "PubSubClient",
        /,
        *,
        exact: Set[bytes],
        glob: Set[str],
        on_receiving: Optional[OnReceiving],
    ) -> None:
        self.state: _PubSubClientSubscriptionState = (
            _PubSubClientSubscriptionStateNotEntered(
                _STATE_NOT_ENTERED, client, exact, glob, on_receiving
            )
        )
        self._state_lock: asyncio.Lock = asyncio.Lock()
        """Protects changing self.state"""

    async def __aenter__(self) -> "PubSubClientSubscription":
        async with self._state_lock:
            state = self.state
            assert state.type == _STATE_NOT_ENTERED, "already entered"
            res = await state.client.direct_subscribe_multiple(
                exact=state.exact,
                glob=state.glob,
            )

            self.state = _PubSubClientSubscriptionStateEnteredNotBuffering(
                type=_STATE_ENTERED_NOT_BUFFERING,
                client=state.client,
                exact=res.exact,
                glob=res.glob,
                on_receiving=state.on_receiving,
            )

        return self

    async def __aexit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc_value: Optional[BaseException],
        traceback: Optional[TracebackType],
    ) -> None:
        async with self._state_lock:
            state = self.state
            if (
                state.type != _STATE_ENTERED_NOT_BUFFERING
                and state.type != _STATE_ENTERED_BUFFERING
            ):
                return

            exc: Optional[BaseException] = None

            if state.type == _STATE_ENTERED_BUFFERING:
                try:
                    await state.client.direct_unregister_on_message(
                        registration_id=state.on_message_subscription_id
                    )
                except BaseException as e:
                    if exc is None:
                        exc = e

                try:
                    await state.client.direct_unregister_status_handler(
                        registration_id=state.on_status_subscription_id
                    )
                except BaseException as e:
                    if exc is None:
                        exc = e

                for message in state.buffer.drain():
                    try:
                        await message.cleanup()
                    except BaseException as e:
                        if exc is None:
                            exc = e

                for message in state.cleanup.drain():
                    try:
                        await message.cleanup()
                    except BaseException as e:
                        if exc is None:
                            exc = e

            for sub_id in state.exact.values():
                try:
                    await state.client.direct_unsubscribe_exact(subscription_id=sub_id)
                except BaseException as e:
                    if exc is None:
                        exc = e

            for sub_id in state.glob.values():
                try:
                    await state.client.direct_unsubscribe_glob(subscription_id=sub_id)
                except BaseException as e:
                    if exc is None:
                        exc = e

            self.state = _PubSubClientSubscriptionStateDisposed(_STATE_DISPOSED)
            if exc is not None:
                raise exc

    async def on_connection_lost(self) -> None:
        # acquiring a state lock leads to a deadlock:
        # - in e.g. aenter, we hold the state lock
        #   - we call direct_subscribe_exact
        #   - this calls either on_connection_lost or on_connection_established
        #
        # luckily, we don't need the state lock so long as we don't yield and are ok
        # with status_queue potentially gaining values while the state lock is held
        if self.state.type == _STATE_ENTERED_BUFFERING:
            self.state.status_queue.put_nowait(PubSubClientConnectionStatus.LOST)

    async def on_connection_established(self) -> None:
        if self.state.type == _STATE_ENTERED_BUFFERING:
            self.state.status_queue.put_nowait(PubSubClientConnectionStatus.OK)

    async def on_connection_abandoned(self) -> None:
        if self.state.type == _STATE_ENTERED_BUFFERING:
            self.state.status_queue.put_nowait(PubSubClientConnectionStatus.ABANDONED)

    async def subscribe_exact(self, topic: bytes) -> None:
        async with self._state_lock:
            state = self.state
            if state.type == _STATE_NOT_ENTERED:
                state.exact.add(topic)
                return

            if state.type == _STATE_DISPOSED:
                raise RuntimeError("subscription has been disposed")

            if topic in state.exact:
                return

            sub_id = await state.client.direct_subscribe_exact(topic=topic)
            state.exact[topic] = sub_id

    async def subscribe_glob(self, glob: str) -> None:
        async with self._state_lock:
            state = self.state
            if state.type == _STATE_NOT_ENTERED:
                state.glob.add(glob)
                return

            if state.type == _STATE_DISPOSED:
                raise RuntimeError("subscription has been disposed")

            if glob in state.glob:
                return

            glob_regex = (
                None
                if state.type != _STATE_ENTERED_BUFFERING
                else re.compile(translate(glob))
            )

            sub_id = await state.client.direct_subscribe_glob(glob=glob)
            state.glob[glob] = sub_id

            if state.type == _STATE_ENTERED_BUFFERING:
                assert glob_regex is not None
                state.glob_list.append(glob)
                state.glob_regexes.append(glob_regex)

    async def unsubscribe_exact(self, topic: bytes) -> None:
        """Unsubscribes from the given exact topic subscription. This is included
        for completeness, but is not necessarily particularly fast.

        Raises ValueError if not subscribed to the given topic, RuntimeError if already
        disposed
        """
        async with self._state_lock:
            state = self.state
            if state.type == _STATE_NOT_ENTERED:
                try:
                    state.exact.remove(topic)
                except KeyError:
                    raise ValueError(f"not subscribed to {topic!r}")
                return

            if state.type == _STATE_DISPOSED:
                raise RuntimeError("subscription has been disposed")

            try:
                sub_id = state.exact.pop(topic)
            except KeyError:
                raise ValueError(f"not subscribed to {topic!r}")

            await state.client.direct_unsubscribe_exact(subscription_id=sub_id)

    async def unsubscribe_glob(self, glob: str) -> None:
        """Unsubscribes from the given glob subscription. This is included
        for completeness, but is not necessarily particularly fast.

        Raises ValueError if not subscribed to the given glob, RuntimeError if already
        disposed
        """
        async with self._state_lock:
            state = self.state
            if state.type == _STATE_NOT_ENTERED:
                try:
                    state.glob.remove(glob)
                except KeyError:
                    raise ValueError(f"not subscribed to {glob!r}")
                return

            if state.type == _STATE_DISPOSED:
                raise RuntimeError("subscription has been disposed")

            try:
                sub_id = state.glob.pop(glob)
            except KeyError:
                raise ValueError(f"not subscribed to {glob!r}")

            exc: Optional[BaseException] = None
            if state.type == _STATE_ENTERED_BUFFERING:
                try:
                    idx = state.glob_list.index(glob)
                    state.glob_list.pop(idx)
                    state.glob_regexes.pop(idx)
                except (IndexError, ValueError) as e:
                    exc = e

            await state.client.direct_unsubscribe_glob(subscription_id=sub_id)

            if exc is not None:
                raise exc

    async def on_message(self, message: PubSubClientMessageWithCleanup) -> None:
        # can avoid a lock by using put_nowait, and raising instead of blocking is
        # preferred in the unlikely event that the queue has a max size AND we reach it
        state = self.state
        if state.type != _STATE_ENTERED_BUFFERING:
            await message.cleanup()
            return

        found = message.topic in state.exact
        if not found:
            try:
                topic_str = message.topic.decode("utf-8")
            except UnicodeDecodeError:
                topic_str = None

            if topic_str is not None:
                for regex in state.glob_regexes:
                    if regex.match(topic_str):
                        found = True
                        break

        if found:
            state.buffer.put_nowait(message)
        else:
            await message.cleanup()

    async def messages(self) -> PubSubClientSubscriptionIterator:
        async with self._state_lock:
            state = self.state
            if state.type == _STATE_NOT_ENTERED:
                raise RuntimeError("not entered")
            if state.type == _STATE_ENTERED_BUFFERING:
                raise RuntimeError("already iterating")
            if state.type == _STATE_DISPOSED:
                raise RuntimeError("subscription has been disposed")
            assert state.type == _STATE_ENTERED_NOT_BUFFERING, "unknown state"

            if (
                state.client.receiver.connection_status
                == PubSubClientConnectionStatus.ABANDONED
            ):
                raise PubSubRequestConnectionAbandonedError("connection abandoned")

            glob_list: List[str] = []
            glob_regexes: List[re.Pattern] = []
            for glob in state.glob.keys():
                glob_list.append(glob)
                glob_regexes.append(re.compile(translate(glob)))

            on_message_subscription_id = await state.client.direct_register_on_message(
                receiver=self
            )
            on_status_subscription_id = (
                await state.client.direct_register_status_handler(receiver=self)
            )

            status = PubSubClientConnectionStatus.LOST
            status_queue: DrainableAsyncioQueue[PubSubClientConnectionStatus] = (
                DrainableAsyncioQueue()
            )
            if state.client.receiver.connection_status != status:
                status_queue.put_nowait(state.client.receiver.connection_status)

            buffer: DrainableAsyncioQueue[PubSubClientMessageWithCleanup] = (
                DrainableAsyncioQueue()
            )
            cleanup: DrainableAsyncioQueue[PubSubClientMessageWithCleanup] = (
                DrainableAsyncioQueue()
            )
            self.state = _PubSubClientSubscriptionStateEnteredBuffering(
                type=_STATE_ENTERED_BUFFERING,
                client=state.client,
                on_message_subscription_id=on_message_subscription_id,
                on_status_subscription_id=on_status_subscription_id,
                status=status,
                status_queue=status_queue,
                exact=state.exact,
                glob_regexes=glob_regexes,
                glob_list=glob_list,
                glob=state.glob,
                buffer=buffer,
                cleanup=cleanup,
                on_receiving=state.on_receiving,
            )
            return PubSubClientSubscriptionIterator(self.state)

    async def with_timeout(
        self, seconds: float
    ) -> PubSubClientSubscriptionWithTimeoutIterator:
        return PubSubClientSubscriptionWithTimeoutIterator(
            await self.messages(), seconds
        )


class PubSubDirectOnMessageWithCleanupReceiver(Protocol):
    async def on_message(self, message: PubSubClientMessageWithCleanup) -> None: ...


class PubSubDirectConnectionStatusReceiver(Protocol):
    """Describes an object that wants to receive feedback about the state of the
    connection, if any information is known. This is most directly interpretable
    with active connections (e.g., websockets)
    """

    async def on_connection_lost(self) -> None:
        """
        Called to indicate that the subscriber knows it's possible that we are
        missing some notifications right now and moving forward. Generally, the
        only thing an implementation can do with this is switch to polling, or in
        practice, set a timeout for an alert if the connection is not established
        again in time.

        NOTE: this will be called under normal circumstances if registering
        status receivers prior to setting up the receiver
        """

    async def on_connection_established(self) -> None:
        """Indicates we (re-)established a connection and now expect that we are
        receiving messages without interruption. This is the most meaningful event
        that implementations can use as it will definitely be called in normal operations
        (when the connection is first established, which, assuming the service regularly
        restarts for e.g. updates, will happen regularly), and by implementating will
        naturally take care of small interruptions in the connection

        There are a few general operations that implementations would perform
        - if using these notifications to fill a local cache, which on misses checks
          the source of truth, just clear the local cache
        - if there is an external log of messages, check for and replay messages
          that weren't processed (taking care to ensure this completes, e.g., by
          marking the current position before starting)
        """

    async def on_connection_abandoned(self) -> None:
        """Indicates the subscriber has given up trying to re-establish the connection and
        will raise errors when trying to receive messages.

        This will happen in normal operation when exiting the context manager for
        the websocket client (i.e., closing the websocket by subscriber request),
        and normally doesn't have any useful recovery. If this was not expected
        and error is going to be raised in the notify/receive methods which can
        be handled with better context
        """


class PubSubError(Exception):
    """Base class for pub/sub exceptions"""


class PubSubIrrecoverableError(BaseException):
    """Base class for pub/sub exceptions that should prevent retries but aren't one
    of the standard base exceptions
    """


class PubSubCancelRequested(PubSubIrrecoverableError):
    """Raised when a cancel is requested"""


class PubSubRequestError(PubSubError):
    """An error occurred while making a request to the pub/sub server"""


class PubSubRequestAmbiguousError(PubSubRequestError):
    """We failed to confirm the server received the request, and we also
    failed to confirm they did not
    """


class PubSubRequestRetriesExhaustedError(PubSubRequestError):
    """Every attempt we made was met with the server explicitly indicating
    we should retry (502, 503, or 504 status code), but we have reached the
    maximum number of retries
    """


class PubSubRequestRefusedError(PubSubRequestError):
    """The server refused the request and indicated we should not retry"""


class PubSubRequestConnectionAbandonedError(PubSubRequestError):
    """We do not have a connection to the broadcaster and we have given up
    trying to establish one
    """


class PubSubNotifyResult(Protocol):
    @property
    def notified(self) -> int:
        """The number of subscribers that were successfully notified. Success means
        either an HTTP 200 response, or, for websocket subscribers, an acknowledgement.
        Ambiguous attempts, such as a connection close after posting the data,
        are never included in this value even though the subscriber may have
        received them
        """


class PubSubClientBulkSubscriptionConnector(Protocol):
    """Something capable of setting the subscriptions in a single call; this
    is an optional extension to the pub sub client connector and is assumed
    to be setup/torn down in the same way as the base connector.

    A connector exposes that they support this via `get_bulk()`
    which returns `self` if its supported and `None` otherwise.

    This incorporates a strong etag so that the subscriptions can be confirmed
    to still be accurate without actually sending the full list of subscriptions
    """

    async def check_subscriptions(self) -> StrongEtag:
        """Retrieves the strong etag describing the subscriptions active

        Returns:
            StrongEtag: the strong etag for the subscriptions
        """

    async def set_subscriptions(
        self,
        /,
        *,
        exact: List[bytes],
        globs: List[str],
    ) -> None:
        """Replaces the subscriptions for the given url with those provided. This
        is assumed to be for a small enough number of topics/globs that holding it
        in memory is reasonable, given the client generally requires that anyway

        Args:
            topics (List[bytes]): the topics to subscribe to
            globs (List[str]): the globs to subscribe to
        """


T_contra = TypeVar("T_contra", contravariant=True)
T_co = TypeVar("T_co", covariant=True)


class PubSubClientTracingNotifyOnHashed(Generic[T_co], Protocol):
    def on_hashed(self) -> T_co: ...


class PubSubClientTracingNotifyStart(Generic[T_co], Protocol):
    def on_start_without_hash(
        self, /, *, topic: bytes, length: int, filelike: bool
    ) -> PubSubClientTracingNotifyOnHashed[T_co]: ...

    def on_start_with_hash(
        self, /, *, topic: bytes, length: int, filelike: bool
    ) -> T_co: ...


NotifierT = TypeVar("NotifierT")
InitializerT = TypeVar("InitializerT")
InitializerTco = TypeVar("InitializerTco", covariant=True)
InitializerTcontra = TypeVar("InitializerTcontra", contravariant=True)


class PubSubClientConnector(Generic[InitializerTcontra, NotifierT], Protocol):
    """Something capable of making subscribe/unsubscribe requests"""

    async def setup_connector(self) -> None:
        """Performs any necessary setup of the connector methods. Must raise an
        error if re-entry is not supported but is attempted
        """

    async def teardown_connector(self) -> None:
        """If called after a setup, must tear down resources created in that setup.
        Otherwise, SHOULD ensure all resources are teared down and MAY raise an
        error.

        If this is implemented such that all resources are as torn down as possible
        after the first call, MAY simply error on subsequent calls
        """

    async def subscribe_exact(self, /, *, topic: bytes) -> None:
        """
        Subscribe to the given topic, such that the corresponding receiver will
        receive one additional notification when a message is posted on that
        topic. MUST raise an error unless this confirmed that during the
        execution of this function the broadcaster had our client subscribed to
        the given topic

        Specifically, this means repeating calls to this function with the same
        topic should typically result in no errors and the same effect as a
        single call.
        """

    async def subscribe_glob(self, /, *, glob: str) -> None:
        """
        Subscribe to the given topic, such that the corresponding receiver will
        receive one additional notification when a message is posted to a matching
        toic. MUST raise an error unless this confirmed that during the
        execution of this function the broadcaster had our client subscribed to
        the given glob

        Specifically, this means repeating calls to this function with the same
        glob should typically result in no errors and the same effect as a
        single call.
        """

    async def unsubscribe_exact(self, /, *, topic: bytes) -> None:
        """
        Unsubscribe from the given topic, such that the corresponding receiver will
        receive one fewer notification when a message is posted on that topic. MUST
        raise an error unless this confirmed that during the execution of this
        function the broadcaster had our client unsubscribed from the given topic

        Specifically, this means repeating calls to this function with the same
        topic should typically result in no errors and the same effect as a
        single call.
        """

    async def unsubscribe_glob(self, /, *, glob: str) -> None:
        """
        Unsubscribe from the given topic, such that the corresponding receiver will
        receive one fewer notification when a message is posted to a matching topic.
        MUST raise an error unless this confirmed that during the execution of this
        function the broadcaster had our client unsubscribed from the given glob

        Specifically, this means repeating calls to this function with the same
        glob should typically result in no errors and the same effect as a
        single call.
        """

    def get_bulk(self) -> Optional[PubSubClientBulkSubscriptionConnector]:
        """Returns a bulk subscription connector if supported, otherwise None"""

    def prepare_notifier_trace(
        self, initializer: InitializerTcontra, /
    ) -> ContextManager[PubSubClientTracingNotifyStart[NotifierT]]:
        """Prepares the tracing object for a notify"""

    async def notify(
        self,
        /,
        *,
        topic: bytes,
        message: SyncStandardIO,
        length: int,
        message_sha512: bytes,
        tracer: NotifierT,
    ) -> PubSubNotifyResult:
        """Sends a message, which is composed of the next length bytes on the given
        seekable synchronous io object, to all subscribers of the given topic.

        MUST raise an error unless this confirmed that during the execution of this
        function the broadcaster received, accepted, and processed the message

        Args:
            topic (bytes): the topic to post the message to
            message (SyncStandardIO): the message to post; the current position is as indicated
                via tell(), and only the next length bytes are part of the message
            length (int): the number of bytes in the message
            message_sha512 (bytes): the sha512 hash of the message (64 bytes)
            tracing (NotifierT): the tracing object to use for this notify
        """


class PubSubClientConnectionStatus(Enum):
    OK = auto()
    """Indicates that we believe we are receiving messages"""

    LOST = auto()
    """Indicates that we believe we may not be receiving messages and are
    trying to re-establish a stable connection
    """

    ABANDONED = auto()
    """Indicates that we believe we may not be receiving messages and are
    not making attempts to rectify the situation
    """


class PubSubClientReceiver(Protocol):
    """Something capable of registering additional callbacks when messages are received"""

    @property
    def connection_status(self) -> PubSubClientConnectionStatus:
        """The subscribers best belief on the current state of the connection. This
        is generally for debugging. See `register_status_handler` for the more useful
        interface programmatically
        """

    async def setup_receiver(self) -> None:
        """Performs any necessary work to prepare to receive messages from the
        broadcaster
        """

    async def teardown_receiver(self) -> None:
        """Called to notify that this object is no longer expected to call status
        handlers or message handlers. Generally, tears down any work done in
        setup_receiver. Implementations MAY assume that after teardown the object
        will not be used again, though in that case they SHOULD raise an error on
        unsupported setup calls.
        """

    async def register_on_message(
        self, /, *, receiver: PubSubDirectOnMessageWithCleanupReceiver
    ) -> int:
        """Registers the given receiver to be called when a message on one of the
        subscribed topics (either via an exact match or a glob match) is received,
        and returns an id that can be used to unregister the receiver.
        """

    async def unregister_on_message(self, /, *, registration_id: int) -> None:
        """Unregisters the receiver with the given id. The implementation MAY assume
        that the registration id is valid (returned from register_on_message from
        this object and not invalidated), and if it is not, may arbitrarily do any
        of the following:
        - corrupt its state
        - unregister an unrelated receiver
        - raise an error
        - do nothing

        Regardless of if this raises an error, afterwards the registration id MUST
        be considered invalidated by the caller
        """

    async def register_status_handler(
        self, /, *, receiver: PubSubDirectConnectionStatusReceiver
    ) -> int:
        """Registers the given receiver to be called when the connection status changes,
        and returns an id that can be used to unregister the receiver.

        The core purpose of the receiver is so that the caller can perform some
        operation when it may have missed messages but is now receiving messages.
        The most general thing it could do is look at a log of the messages sent
        over the topic stored elsewhere and replay any it hasn't seen, but most of
        the time there is a simpler alternative.

        For example, if the caller does the same idempotent operation regardless
        of the contents/topic of the message, then they can simply do that operation
        on `on_connection_established`
        """

    async def unregister_status_handler(self, /, *, registration_id: int) -> None:
        """Unregisters the receiver with the given id. The implementation MAY assume
        that the registration id is valid (returned from register_status_handler from
        this object and not invalidated), and if it is not, may arbitrarily do any
        of the following:
        - corrupt its state
        - unregister an unrelated receiver
        - raise an error
        - do nothing

        Regardless of if this raises an error, afterwards the registration id MUST
        be considered invalidated by the caller
        """


@fast_dataclass
class DirectSubscribeMultipleResult:
    """The result of PubSubClient#direct_subscribe_multiple"""

    glob: Dict[str, int]
    """the glob subscriptions requested mapped to the id to unregister
    the corresponding subscription
    """

    exact: Dict[bytes, int]
    """the exact subscriptions requested mapped to the id to unregister
    the corresponding subscription
    """


COST_PER_INDIV_REQ = 2**16
"""The cost of making an individual request; unitless"""

COST_PER_BULK_ITEM = 128
"""The cost per each item within a bulk request; unitless"""


class PubSubClient(Generic[InitializerT, NotifierT]):
    def __init__(
        self,
        connector: PubSubClientConnector[InitializerT, NotifierT],
        receiver: PubSubClientReceiver,
        *,
        setup: Callable[[], Awaitable[None]],
        teardown: Callable[[], Awaitable[None]],
    ) -> None:
        self.connector: PubSubClientConnector = connector
        """The connector that can make subscribe/unsubscribe requests. We assume that we
        need to setup this when we are entered and teardown when we are exited.
        """

        self.receiver: PubSubClientReceiver = receiver
        """The receiver that can register additional callbacks when messages are received.
        We assume that we need to setup this when we are entered and teardown when we are exited.
        """

        self._setup = setup
        """A function to call when we are entered"""

        self._teardown = teardown
        """A function to call when we are exited"""

        self.exact_subscriptions: Dict[bytes, int] = {}
        """Maps from topic we've subscribed to to the number of requests to subscribe to it"""

        self.active_exact_subscriptions: Dict[int, bytes] = {}
        """Maps from subscription_id to the corresponding exact topic"""

        self.glob_subscriptions: Dict[str, int] = {}
        """Maps from glob we've subscribed to to the number of requests to subscribe to it"""

        self.active_glob_subscriptions: Dict[int, str] = {}
        """Maps from subscription_id to the corresponding glob"""

        self._entered: bool = False
        """True if we are active (aenter without aexit), False otherwise"""

        self._counter: int = 0
        """The counter for generating unique subscription ids"""

        self._subscribing_lock: asyncio.Lock = asyncio.Lock()
        """A lock while actively subscribing/unsubscribing; managed by the direct_*
        methods
        """

    async def __aenter__(self) -> "PubSubClient":
        assert not self._entered, "already entered (not re-entrant)"
        await self._setup()
        self._entered = True
        try:
            await self.connector.setup_connector()
            try:
                await self.receiver.setup_receiver()
            except BaseException:
                await self.connector.teardown_connector()
                raise
        except BaseException:
            try:
                await self._teardown()
            finally:
                self._entered = False

            raise

        return self

    async def __aexit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc_value: Optional[BaseException],
        traceback: Optional[TracebackType],
    ) -> None:
        assert self._entered, "not entered"

        self._entered = False
        excs: List[BaseException] = []

        try:
            await self.receiver.teardown_receiver()
        except BaseException as e:
            excs.append(e)

        async with self._subscribing_lock:
            if (bulk := self.connector.get_bulk()) is not None and (
                self.exact_subscriptions or self.glob_subscriptions
            ):
                try:
                    await bulk.set_subscriptions(exact=[], globs=[])
                    self.exact_subscriptions = dict()
                    self.active_exact_subscriptions = dict()
                    self.glob_subscriptions = dict()
                    self.active_glob_subscriptions = dict()
                except BaseException as e:
                    excs.append(e)

            for topic in self.exact_subscriptions.keys():
                try:
                    await self.connector.unsubscribe_exact(topic=topic)
                except BaseException as e:
                    excs.append(e)
            for glob in self.glob_subscriptions.keys():
                try:
                    await self.connector.unsubscribe_glob(glob=glob)
                except BaseException as e:
                    excs.append(e)

            self.exact_subscriptions = dict()
            self.active_exact_subscriptions = dict()
            self.glob_subscriptions = dict()
            self.active_glob_subscriptions = dict()

        try:
            await self.connector.teardown_connector()
        except BaseException as e:
            excs.append(e)

        try:
            await self._teardown()
        except BaseException as e:
            excs.append(e)

        if excs:
            raise combine_multiple_exceptions("failed to teardown", excs)

    def _reserve_subscription_id(self) -> int:
        """An asyncio-safe (since it doesn't yield) way to reserve a subscription
        id. Not thread-safe
        """
        result = self._counter
        self._counter += 1
        return result

    async def _try_direct_subscribe_exact(
        self, /, *, topic: bytes, my_id: int, have_lock: bool
    ) -> Literal["ok", "need_lock"]:
        requests_so_far = self.exact_subscriptions.get(topic, 0)
        if requests_so_far <= 0:
            if not have_lock:
                return "need_lock"

            await self.connector.subscribe_exact(topic=topic)

        self.exact_subscriptions[topic] = max(1, requests_so_far + 1)
        self.active_exact_subscriptions[my_id] = topic
        return "ok"

    async def _direct_subscribe_exact_via_individual(self, /, *, topic: bytes) -> int:
        assert self._entered, "not entered"
        my_id = self._reserve_subscription_id()
        result = await self._try_direct_subscribe_exact(
            topic=topic, my_id=my_id, have_lock=False
        )
        if result == "need_lock":
            async with self._subscribing_lock:
                result = await self._try_direct_subscribe_exact(
                    topic=topic, my_id=my_id, have_lock=True
                )
        assert result == "ok"
        return my_id

    async def _try_direct_subscribe_glob(
        self, /, *, glob: str, my_id: int, have_lock: bool
    ) -> Literal["ok", "need_lock"]:
        requests_so_far = self.glob_subscriptions.get(glob, 0)
        if requests_so_far == 0:
            if not have_lock:
                return "need_lock"

            await self.connector.subscribe_glob(glob=glob)

        self.glob_subscriptions[glob] = max(1, requests_so_far + 1)
        self.active_glob_subscriptions[my_id] = glob
        return "ok"

    async def _direct_subscribe_glob_via_individual(self, /, *, glob: str) -> int:
        assert self._entered, "not entered"
        my_id = self._reserve_subscription_id()
        result = await self._try_direct_subscribe_glob(
            glob=glob, my_id=my_id, have_lock=False
        )
        if result == "need_lock":
            async with self._subscribing_lock:
                result = await self._try_direct_subscribe_glob(
                    glob=glob, my_id=my_id, have_lock=True
                )
        assert result == "ok"
        return my_id

    async def _try_direct_unsubscribe_exact(
        self, /, *, subscription_id: int, have_lock: bool
    ) -> Literal["ok", "need_lock"]:
        topic = self.active_exact_subscriptions.get(subscription_id)
        if topic is None:
            return "ok"

        requests_so_far = self.exact_subscriptions[topic]
        need_unsubscribe = requests_so_far <= 1
        if need_unsubscribe and not have_lock:
            return "need_lock"

        # If we're unsubscribing we'll set the value in exact_subscriptions to 0
        # temporarily, then remove it if we're successful. The only effect of this
        # is that we might want to use this information when we aexit
        del self.active_exact_subscriptions[subscription_id]
        self.exact_subscriptions[topic] = requests_so_far - 1

        if need_unsubscribe:
            await self.connector.unsubscribe_exact(topic=topic)
            assert self.exact_subscriptions[topic] <= 0
            del self.exact_subscriptions[topic]

        return "ok"

    async def _direct_unsubscribe_exact_via_individual(
        self, /, *, subscription_id: int
    ) -> None:
        assert self._entered, "not entered"
        result = await self._try_direct_unsubscribe_exact(
            subscription_id=subscription_id, have_lock=False
        )
        if result == "need_lock":
            async with self._subscribing_lock:
                result = await self._try_direct_unsubscribe_exact(
                    subscription_id=subscription_id, have_lock=True
                )
        assert result == "ok"

    async def _direct_set_subscriptions(
        self, bulk: PubSubClientBulkSubscriptionConnector
    ) -> None:
        """sets the subscriptions to what we expect; should have the lock"""
        await bulk.set_subscriptions(
            exact=list(k for k, v in self.exact_subscriptions.items() if v > 0),
            globs=list(k for k, v in self.glob_subscriptions.items() if v > 0),
        )

        topics_to_rem: List[bytes] = []
        for topic, num_requests in self.exact_subscriptions.items():
            if num_requests <= 0:
                topics_to_rem.append(topic)
        for topic in topics_to_rem:
            del self.exact_subscriptions[topic]

        globs_to_rem: List[str] = []
        for glob, num_requests in self.glob_subscriptions.items():
            if num_requests <= 0:
                globs_to_rem.append(glob)
        for glob in globs_to_rem:
            del self.glob_subscriptions[glob]

    async def _try_direct_unsubscribe_exact_via_bulk(
        self, /, *, subscription_id: int, have_lock: bool
    ) -> Literal["ok", "unsupported", "not_desirable", "need_lock"]:
        assert self._entered, "not entered"

        topic = self.active_exact_subscriptions.get(subscription_id)
        if topic is None:
            return "ok"

        num_requests = self.exact_subscriptions.get(topic, 0)
        if num_requests > 1:
            del self.active_exact_subscriptions[subscription_id]
            self.exact_subscriptions[topic] = num_requests - 1
            return "ok"

        if (bulk := self.connector.get_bulk()) is None:
            return "unsupported"

        cost_for_individual = COST_PER_INDIV_REQ
        cost_for_bulk = COST_PER_BULK_ITEM * (
            (len(self.active_exact_subscriptions) - 1)
            + len(self.active_glob_subscriptions)
        )
        if cost_for_individual < cost_for_bulk:
            return "not_desirable"

        if not have_lock:
            return "need_lock"

        del self.active_exact_subscriptions[subscription_id]
        self.exact_subscriptions[topic] = 0

        await self._direct_set_subscriptions(bulk)
        return "ok"

    async def _try_direct_unsubscribe_glob(
        self, /, *, subscription_id: int, have_lock: bool
    ) -> Literal["ok", "need_lock"]:
        glob = self.active_glob_subscriptions.get(subscription_id)
        if glob is None:
            return "ok"

        requests_so_far = self.glob_subscriptions[glob]
        need_unsubscribe = requests_so_far <= 1
        if need_unsubscribe and not have_lock:
            return "need_lock"

        del self.active_glob_subscriptions[subscription_id]
        self.glob_subscriptions[glob] = requests_so_far - 1

        if need_unsubscribe:
            await self.connector.unsubscribe_glob(glob=glob)
            assert self.glob_subscriptions[glob] <= 0
            del self.glob_subscriptions[glob]

        return "ok"

    async def _direct_unsubscribe_glob_via_individual(
        self, /, *, subscription_id: int
    ) -> None:
        assert self._entered, "not entered"
        result = await self._try_direct_unsubscribe_glob(
            subscription_id=subscription_id, have_lock=False
        )
        if result == "need_lock":
            async with self._subscribing_lock:
                result = await self._try_direct_unsubscribe_glob(
                    subscription_id=subscription_id, have_lock=True
                )
        assert result == "ok"

    async def _try_direct_unsubscribe_glob_via_bulk(
        self, /, *, subscription_id: int, have_lock: bool
    ) -> Literal["ok", "unsupported", "not_desirable", "need_lock"]:
        assert self._entered, "not entered"

        glob = self.active_glob_subscriptions.get(subscription_id)
        if glob is None:
            return "ok"

        num_requests = self.glob_subscriptions.get(glob, 0)
        if num_requests > 1:
            del self.active_glob_subscriptions[subscription_id]
            self.glob_subscriptions[glob] = num_requests - 1
            return "ok"

        if (bulk := self.connector.get_bulk()) is None:
            return "unsupported"

        cost_for_individual = COST_PER_INDIV_REQ
        cost_for_bulk = COST_PER_BULK_ITEM * (
            len(self.active_exact_subscriptions)
            + (len(self.active_glob_subscriptions) - 1)
        )
        if cost_for_individual < cost_for_bulk:
            return "not_desirable"

        if not have_lock:
            return "need_lock"

        del self.active_glob_subscriptions[subscription_id]
        self.glob_subscriptions[glob] = 0

        await self._direct_set_subscriptions(bulk)
        return "ok"

    async def _try_direct_subscribe_multiple_via_individual_calls(
        self,
        /,
        *,
        exact: Iterable[bytes],
        glob: Iterable[str],
    ) -> DirectSubscribeMultipleResult:
        result = DirectSubscribeMultipleResult(
            exact=dict(),
            glob=dict(),
        )
        try:
            for topic in exact:
                result.exact[topic] = await self._direct_subscribe_exact_via_individual(
                    topic=topic
                )
            for gb in glob:
                result.glob[gb] = await self._direct_subscribe_glob_via_individual(
                    glob=gb
                )
            return result
        except BaseException as context:
            unrelated: List[BaseException] = []

            for sub_id in result.exact.values():
                try:
                    await self._direct_unsubscribe_exact_via_individual(
                        subscription_id=sub_id
                    )
                except BaseException as e:
                    unrelated.append(e)
            for sub_id in result.glob.values():
                try:
                    await self._direct_unsubscribe_glob_via_individual(
                        subscription_id=sub_id
                    )
                except BaseException as e:
                    unrelated.append(e)

            raise combine_multiple_exceptions(
                "failed to subscribe via multiple subscribe calls",
                unrelated,
                context=context,
            )

    async def _try_direct_subscribe_multiple_via_bulk_call_if_efficient(
        self,
        /,
        *,
        exact: Iterable[bytes],
        glob: Iterable[str],
    ) -> Optional[DirectSubscribeMultipleResult]:
        bulk = self.connector.get_bulk()
        if bulk is None:
            return None

        to_add = 0
        for topic in exact:
            if self.exact_subscriptions.get(topic, 0) <= 0:
                to_add += 1

        for gb in glob:
            if self.glob_subscriptions.get(gb, 0) <= 0:
                to_add += 1

        cost_bulk = COST_PER_BULK_ITEM * (
            len(self.active_exact_subscriptions)
            + len(self.active_glob_subscriptions)
            + to_add
        )
        cost_indiv = COST_PER_INDIV_REQ * to_add

        if cost_indiv < cost_bulk:
            return None

        async with self._subscribing_lock:
            result = DirectSubscribeMultipleResult(glob=dict(), exact=dict())

            for topic in exact:
                sub_id = self._reserve_subscription_id()
                self.active_exact_subscriptions[sub_id] = topic
                result.exact[topic] = sub_id
                self.exact_subscriptions[topic] = (
                    max(self.exact_subscriptions.get(topic, 0), 0) + 1
                )

            for glob in glob:
                sub_id = self._reserve_subscription_id()
                self.active_glob_subscriptions[sub_id] = glob
                result.glob[glob] = sub_id
                self.glob_subscriptions[glob] = (
                    max(self.glob_subscriptions.get(glob, 0), 0) + 1
                )

            await self._direct_set_subscriptions(bulk)
            return result

    async def direct_subscribe_multiple(
        self,
        /,
        exact: Iterable[bytes],
        glob: Iterable[str],
    ) -> DirectSubscribeMultipleResult:
        """Subscribes to multiple topics and glob patterns at the same time,
        returning the corresponding ids for unregistering the subscriptions.
        This may be more efficient or more resilient than multiple calls to
        the individual methods, depending on the connection type.

        Args:
            exact (Iterable[bytes]): the topics to subscribe to
            glob (Iterable[str]): the glob patterns to subscribe to

        Returns:
            DirectSubscribeMultipleResult: the ids to unregister the subscriptions
        """
        result = await self._try_direct_subscribe_multiple_via_bulk_call_if_efficient(
            exact=exact, glob=glob
        )
        if result is None:
            result = await self._try_direct_subscribe_multiple_via_individual_calls(
                exact=exact, glob=glob
            )
        return result

    async def direct_subscribe_exact(self, /, *, topic: bytes) -> int:
        """If we are not already subscribed to the given topic, subscribe to it.
        Returns an id that must be provided to `direct_unsubscribe_exact` when
        the caller is no longer interested in the topic. The caller should register
        with `direct_register_on_message` to receive messages, filtering to
        those it cares about.

        WARN:
            prefer using the `subscribe*` methods instead, which will handle
            unsubscribing via an async context manager. otherwise, cleanup is
            both tedious and error-prone

        WARN:
            the returned id is only guarranteed to be unique to other _active_
            subscriptions. we may reuse values once they are no longer active
            (via `direct_unsubscribe_exact`)

        WARN:
            on error it may be ambiguous if we are subscribed or not
        """
        return (await self.direct_subscribe_multiple(exact=[topic], glob=[])).exact[
            topic
        ]

    async def direct_subscribe_glob(self, /, *, glob: str) -> int:
        """If we are not already subscribed to the given glob, subscribe to it.
        Returns an id that must be provided to `direct_unsubscribe_glob` when
        the caller is no longer interested in the topic. The caller should register
        with `direct_register_on_message` to receive messages, filtering to
        those it cares about.

        WARN:
            prefer using the `subscribe*` methods instead, which will handle
            unsubscribing via an async context manager. otherwise, cleanup is
            both tedious and error-prone

        WARN:
            the returned id is only guarranteed to be unique to other _active_
            subscriptions. we may reuse values once they are no longer active
            (via `direct_unsubscribe_glob`)

        WARN:
            if the glob overlaps with another glob or exact subscription, we
            will receive multiple messages for the same topic with no way to
            deduplicate them (you can put a message uid in the body to detect
            them). this is an intentional limitation as deduplicating can be
            very expensive, this type of duplication can usually be designed
            around, and duplication from network errors needs to be handled
            anyway so should not cause logic errors
        """
        return (await self.direct_subscribe_multiple(exact=[], glob=[glob])).glob[glob]

    async def direct_unsubscribe_exact(self, /, *, subscription_id: int) -> None:
        """If the subscription id was returned from `direct_subscribe_exact`, and
        it has not already been unsubscribed via this method, then unsubscribe
        from the topic. If the subscription id is not as indicated, the behavior
        is undefined:
        - it may do nothing
        - it may raise an error
        - it may unsubscribe an unrelated subscription
        - it may corrupt the state of the client

        WARN:
            prefer using the `subscribe*` methods instead, which will handle
            unsubscribing via an async context manager. otherwise, cleanup is
            both tedious and error-prone
        """
        result = await self._try_direct_unsubscribe_exact_via_bulk(
            subscription_id=subscription_id, have_lock=False
        )
        if result == "ok":
            return

        async with self._subscribing_lock:
            result = await self._try_direct_unsubscribe_exact_via_bulk(
                subscription_id=subscription_id, have_lock=True
            )
            if result != "ok":
                result = await self._try_direct_unsubscribe_exact(
                    subscription_id=subscription_id, have_lock=True
                )
            assert result == "ok"

    async def direct_unsubscribe_glob(self, /, *, subscription_id: int) -> None:
        """If the subscription id was returned from `direct_subscribe_glob`, and
        it has not already been unsubscribed via this method, then unsubscribe
        from the topic. If the subscription id is not as indicated, the behavior
        is undefined:
        - it may do nothing
        - it may raise an error
        - it may unsubscribe an unrelated subscription
        - it may corrupt the state of the client

        WARN:
            prefer using the `subscribe*` methods instead, which will handle
            unsubscribing via an async context manager. otherwise, cleanup is
            both tedious and error-prone
        """
        result = await self._try_direct_unsubscribe_glob_via_bulk(
            subscription_id=subscription_id, have_lock=False
        )
        if result == "ok":
            return

        async with self._subscribing_lock:
            result = await self._try_direct_unsubscribe_glob_via_bulk(
                subscription_id=subscription_id, have_lock=True
            )
            if result != "ok":
                result = await self._try_direct_unsubscribe_glob(
                    subscription_id=subscription_id, have_lock=True
                )
            assert result == "ok"

    async def direct_register_on_message(
        self, /, *, receiver: PubSubDirectOnMessageWithCleanupReceiver
    ) -> int:
        """Registers the given callback to be invoked whenever we receive a message
        for any topic. Returns a registration id that must be provided to
        `direct_unregister_on_message` when the caller is no longer interested in
        the messages.

        WARN:
            prefer using the `subscribe*` methods instead, which will handle
            unsubscribing via an async context manager. otherwise, cleanup is
            both tedious and error-prone
        """
        assert self._entered, "not entered"
        return await self.receiver.register_on_message(receiver=receiver)

    async def direct_unregister_on_message(self, /, *, registration_id: int) -> None:
        """If the registration id was returned from `direct_register_on_message`, and
        it has not already been unregistered via this method, then unregister the
        callback. If the registration id is not as indicated, the behavior is undefined:
        - it may do nothing
        - it may raise an error
        - it may unregister an unrelated callback
        - it may corrupt the state of the client

        WARN:
            prefer using the `subscribe*` methods instead, which will handle
            unsubscribing via an async context manager. otherwise, cleanup is
            both tedious and error-prone
        """
        assert self._entered, "not entered"
        return await self.receiver.unregister_on_message(
            registration_id=registration_id
        )

    async def direct_register_status_handler(
        self, /, *, receiver: PubSubDirectConnectionStatusReceiver
    ) -> int:
        """Registers the given callback to be invoked whenever we receive a connection
        status update. Returns a registration id that must be provided to
        `direct_unregister_status_handler` when the caller is no longer interested in
        the messages.

        WARN:
            the subscribe* methods will handle registering a status handler for you
            while also checking the state of the connection, and thus should be preferred
        """
        assert self._entered, "not entered"
        return await self.receiver.register_status_handler(receiver=receiver)

    async def direct_unregister_status_handler(
        self, /, *, registration_id: int
    ) -> None:
        """If the registration id was returned from `direct_register_status_handler`, and
        it has not already been unregistered via this method, then unregister the
        callback. If the registration id is not as indicated, the behavior is undefined:
        - it may do nothing
        - it may raise an error
        - it may unregister an unrelated callback
        - it may corrupt the state of the client

        WARN:
            prefer using the `subscribe*` methods instead, which will handle
            unsubscribing via an async context manager. otherwise, cleanup is
            both tedious and error-prone
        """
        assert self._entered, "not entered"
        return await self.receiver.unregister_status_handler(
            registration_id=registration_id
        )

    def subscribe_multi(
        self, *, on_receiving: Optional[OnReceiving] = None
    ) -> PubSubClientSubscription:
        """Returns a new async context manager within which you can
        register multiple subscriptions (exact or glob). When exiting,
        the subscriptions will be removed.

        If `on_receiving` is provided, it will be called after any period where
        we may not have been receiving messages and we are likely now receiving
        messages, after buffering begins. This is always called at least once when
        making the iterator (assuming no errors).

        Re-entrant subscriptions are supported and avoid duplicate subscribe/
        unsubscribe requests to the broadcaster

        All the other `subscribe_*` methods behave similarly to this but with
        a bit of setup in advance.
        """
        assert self._entered, "not entered"
        return PubSubClientSubscription(
            self, exact=set(), glob=set(), on_receiving=on_receiving
        )

    def subscribe_exact(
        self, topic: bytes, *rest: bytes, on_receiving: Optional[OnReceiving] = None
    ) -> PubSubClientSubscription:
        """Subscribe to one or more topics by exact match. The result is an
        async context manager which, when exited, will unsubscribe from the
        topic(s)

        If `on_receiving` is provided, it will be called after any period where
        we may not have been receiving messages and we are likely now receiving
        messages, after buffering begins. This is always called at least once when
        making the iterator (assuming no errors).

        Re-entrant subscriptions are supported and avoid duplicate subscribe/
        unsubscribe requests to the broadcaster
        """
        assert self._entered, "not entered"
        return PubSubClientSubscription(
            self, exact={topic, *rest}, glob=set(), on_receiving=on_receiving
        )

    def subscribe_glob(
        self, glob: str, *rest: str, on_receiving: Optional[OnReceiving] = None
    ) -> PubSubClientSubscription:
        """Subscribe to one or more topics by glob match. The result is an
        async context manager which, when exited, will unsubscribe from the
        topic(s)

        If `on_receiving` is provided, it will be called after any period where
        we may not have been receiving messages and we are likely now receiving
        messages, after buffering begins. This is always called at least once when
        making the iterator (assuming no errors).

        Re-entrant subscriptions are supported and avoid duplicate subscribe/
        unsubscribe requests to the broadcaster
        """
        assert self._entered, "not entered"
        return PubSubClientSubscription(
            self, exact=set(), glob={glob, *rest}, on_receiving=on_receiving
        )

    def subscribe(
        self,
        /,
        *,
        glob: Optional[Iterable[str]] = None,
        exact: Optional[Iterable[bytes]] = None,
        on_receiving: Optional[OnReceiving] = None,
    ) -> PubSubClientSubscription:
        """Subscribe to a combination of glob and/or exact topics. The result is
        an async context manager which, when exited, will unsubscribe from the
        topic(s)

        If `on_receiving` is provided, it will be called after any period where
        we may not have been receiving messages and we are likely now receiving
        messages, after buffering begins. This is always called at least once when
        making the iterator (assuming no errors).

        Re-entrant subscriptions are supported and avoid duplicate subscribe/
        unsubscribe requests to the broadcaster
        """
        assert self._entered, "not entered"
        return PubSubClientSubscription(
            self,
            exact=set(exact or ()),
            glob=set(glob or ()),
            on_receiving=on_receiving,
        )

    @overload
    async def notify(
        self,
        /,
        *,
        trace_initializer: InitializerT,
        topic: bytes,
        data: bytes,
        sha512: Optional[bytes] = None,
    ) -> PubSubNotifyResult: ...

    @overload
    async def notify(
        self,
        /,
        *,
        trace_initializer: InitializerT,
        topic: bytes,
        sync_file: SyncStandardIO,
        length: Optional[int] = None,
        sha512: Optional[bytes] = None,
    ) -> PubSubNotifyResult: ...

    async def notify(
        self,
        /,
        *,
        trace_initializer: InitializerT,
        topic: bytes,
        data: Optional[bytes] = None,
        sync_file: Optional[SyncStandardIO] = None,
        length: Optional[int] = None,
        sha512: Optional[bytes] = None,
    ) -> PubSubNotifyResult:
        """Notifies all subscribers of the given topic of the message. The message
        may be provided as bytes or a readable synchronous file-like object.

        If the sha512 is not provided it will be calculated from the message, which
        will incidentally discover the length of the message (via seeking to EOF)
        if not provided. If the length and sha512 are provided then the file
        will only be read once.

        If the message is provided as bytes and length is set, then the length must
        be equal to len(message).
        """
        assert (
            data is None or sync_file is None
        ), "only one of data or sync_file may be provided"
        assert (
            data is not None or sync_file is not None
        ), "either data or sync_file must be provided"
        assert (
            data is None or length is None or len(data) == length
        ), "if data is provided, length must be None or len(data)"
        assert len(topic) <= 65535, "topic too long"
        assert self._entered, "not entered"

        with self.connector.prepare_notifier_trace(
            trace_initializer
        ) as on_start_tracer:
            if sync_file is not None:
                file_starts_at = sync_file.tell()
                if length is None:
                    length = sync_file.seek(0, os.SEEK_END) - file_starts_at
                sync_file = PositionedSyncStandardIO(sync_file, file_starts_at, length)
                del file_starts_at
                sync_file.seek(0, os.SEEK_SET)
            else:
                assert data is not None, "impossible"
                length = len(data)
                sync_file = BytesIO(data)

            # message is used to make it clear to the type checker that either
            # data or sync_file is set

            if sha512 is None:
                on_hashed_tracer = on_start_tracer.on_start_without_hash(
                    topic=topic, length=length, filelike=data is None
                )
                del on_start_tracer

                if data is not None:
                    sha512 = hashlib.sha512(data).digest()
                else:
                    hasher = hashlib.sha512()
                    while True:
                        chunk = sync_file.read(8192)
                        if not chunk:
                            break
                        hasher.update(chunk)
                        await asyncio.sleep(0)
                    sha512 = hasher.digest()
                    sync_file.seek(0, os.SEEK_SET)

                on_send_tracer = on_hashed_tracer.on_hashed()
            else:
                on_send_tracer = on_start_tracer.on_start_with_hash(
                    topic=topic, length=length, filelike=data is None
                )

            return await self.connector.notify(
                topic=topic,
                message=sync_file,
                length=length,
                message_sha512=sha512,
                tracer=on_send_tracer,
            )



import asyncio
from typing import Generic, Protocol, TypeVar

from lonelypsp.util.bounded_deque import BoundedDequeFullError

from lonelypsc.client import PubSubIrrecoverableError
from lonelypsc.util.errors import combine_multiple_exceptions
from lonelypsc.ws.state import State

T = TypeVar("T")
T_contra = TypeVar("T_contra", contravariant=True)


class GenericStateHandler(Generic[T_contra], Protocol):
    """Like a state handler but with the incoming state type already restricted"""

    async def __call__(self, state: T_contra, /) -> State: ...


class RecoveryStateHandler(Generic[T_contra], Protocol):
    """Like a state handler but given the cause that it was run"""

    async def __call__(self, state: T_contra, /, *, cause: BaseException) -> State: ...


async def handle_via_composition(
    state: T,
    *,
    core: GenericStateHandler[T],
    recover: RecoveryStateHandler[T],
    shutdown: RecoveryStateHandler[T],
) -> State:
    """Returns the new state to move to using the standard error handling
    approach.

    Args:
        core (StateHandler for T): the handler that will return the happy path state
            to move to or raise an error
        recovery (StateHandler with exception for T): the handler that will
            cleanup any state-specific resources optimistically such that it can
            move onto the next state, raising an error if it is now
            irrecoverable
        shutdown (StateHandler with exception  for T): the handler that will cleanup
            any state-specific resources pessimistically and eventually bubble the error
            up (possible after going through CLOSING)

    Returns:
        the next state
    """
    exception: BaseException
    try:
        return await core(state)
    except (
        NotImplementedError,
        AssertionError,
        PubSubIrrecoverableError,
        asyncio.CancelledError,
        KeyboardInterrupt,
        # full queues typically cause dropped state, e.g., what we are meant
        # to be subscribed to, which can't be recovered from
        BoundedDequeFullError,
        asyncio.QueueFull,
    ) as e:
        recoverable = False
        exception = e
    except BaseException as be:
        recoverable = True
        exception = be

    if recoverable:
        try:
            return await recover(state, cause=exception)
        except BaseException as be:
            recoverable = False
            exception = combine_multiple_exceptions(
                "while recovering from state", [be], context=exception
            )

    return await shutdown(state, cause=exception)
